"""Installer CLI."""

import argparse
import os.path
import sys
import sysconfig
from typing import Dict, Optional, Sequence

import installer
from installer.destinations import SchemeDictionaryDestination
from installer.sources import WheelFile
from installer.utils import get_launcher_kind


def _get_main_parser() -> argparse.ArgumentParser:
    """Construct the main parser."""
    parser = argparse.ArgumentParser()
    parser.add_argument("wheel", type=str, help="wheel file to install")
    parser.add_argument(
        "--destdir",
        "-d",
        metavar="path",
        type=str,
        help="destination directory (prefix to prepend to each file)",
    )
    parser.add_argument(
        "--prefix",
        "-p",
        metavar="path",
        type=str,
        help="override prefix to install packages to",
    )
    parser.add_argument(
        "--compile-bytecode",
        action="append",
        metavar="level",
        type=int,
        choices=[0, 1, 2],
        help="generate bytecode for the specified optimization level(s) (default=0, 1)",
    )
    parser.add_argument(
        "--no-compile-bytecode",
        action="store_true",
        help="don't generate bytecode for installed modules",
    )
    return parser


def _get_scheme_dict(
    distribution_name: str, prefix: Optional[str] = None
) -> Dict[str, str]:
    """Calculate the scheme dictionary for the current Python environment."""
    vars = {}
    if prefix is None:
        installed_base = sysconfig.get_config_var("base")
        assert installed_base
    else:
        vars["base"] = vars["platbase"] = installed_base = prefix

    scheme_dict = sysconfig.get_paths(vars=vars)

    # calculate 'headers' path, not currently in sysconfig - see
    # https://bugs.python.org/issue44445. This is based on what distutils does.
    # TODO: figure out original vs normalised distribution names
    scheme_dict["headers"] = os.path.join(
        sysconfig.get_path("include", vars={"installed_base": installed_base}),
        distribution_name,
    )

    return scheme_dict


def _main(cli_args: Sequence[str], program: Optional[str] = None) -> None:
    """Process arguments and perform the install."""
    parser = _get_main_parser()
    if program:
        parser.prog = program
    args = parser.parse_args(cli_args)

    bytecode_levels = args.compile_bytecode
    if args.no_compile_bytecode:
        bytecode_levels = []
    elif not bytecode_levels:
        bytecode_levels = [0, 1]

    with WheelFile.open(args.wheel) as source:
        destination = SchemeDictionaryDestination(
            scheme_dict=_get_scheme_dict(source.distribution, prefix=args.prefix),
            interpreter=sys.executable,
            script_kind=get_launcher_kind(),
            bytecode_optimization_levels=bytecode_levels,
            destdir=args.destdir,
        )
        installer.install(source, destination, {})


if __name__ == "__main__":  # pragma: no cover
    _main(sys.argv[1:], "python -m installer")



# Copyright 2012-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import asyncio

import portage
from portage import os
from portage.dbapi.porttree import _async_manifest_fetchlist
from portage.dep import _repo_separator
from portage.localization import _
from portage.util._async.AsyncScheduler import AsyncScheduler
from .ManifestTask import ManifestTask


async def manifest_scheduler_retry(*args, max_tries=3, **kwargs) -> int:
    """
    Since ManifestScheduler is not well suited for internal retry, create
    a new ManifestScheduler instance for each retry. Returns the returncode
    from the last ManifestScheduler instance, which will only be non-zero
    if all retries failed.
    """
    tries = max_tries
    scheduler = ManifestScheduler(*args, **kwargs)
    scheduler.start()
    try:
        await scheduler.async_wait()
    except asyncio.CancelledError:
        scheduler.terminate()
        await scheduler.async_wait()
        raise
    cp_failed = scheduler.cp_failed.copy()
    tries -= 1
    while scheduler.cp_retry and tries > 0:
        kwargs["cp_iter"] = iter(scheduler.cp_retry)
        scheduler = ManifestScheduler(*args, **kwargs)
        scheduler.start()
        try:
            await scheduler.async_wait()
        except asyncio.CancelledError:
            scheduler.terminate()
            await scheduler.async_wait()
            raise
        cp_failed.update(scheduler.cp_failed)
        cp_failed.difference_update(scheduler.cp_successful)
        tries -= 1

    # Account for failures from all ManifestScheduler instances, since we
    # only retry when the returncode is unexpected.
    if cp_failed:
        scheduler.returncode |= 1

    return scheduler.returncode


class ManifestScheduler(AsyncScheduler):
    def __init__(
        self,
        portdb,
        cp_iter=None,
        gpg_cmd=None,
        gpg_vars=None,
        force_sign_key=None,
        **kwargs,
    ):
        AsyncScheduler.__init__(self, **kwargs)

        self._portdb = portdb

        if cp_iter is None:
            cp_iter = self._iter_every_cp()
        self._cp_iter = cp_iter
        self._gpg_cmd = gpg_cmd
        self._gpg_vars = gpg_vars
        self._force_sign_key = force_sign_key
        self._task_iter = self._iter_tasks()
        self.cp_retry = set()
        self.cp_failed = set()
        self.cp_successful = set()

    def _next_task(self):
        return next(self._task_iter)

    def _iter_every_cp(self):
        # List categories individually, in order to start yielding quicker,
        # and in order to reduce latency in case of a signal interrupt.
        cp_all = self._portdb.cp_all
        for category in sorted(self._portdb.categories):
            yield from cp_all(categories=(category,))

    def _iter_tasks(self):
        portdb = self._portdb
        distdir = portdb.settings["DISTDIR"]
        disabled_repos = set()

        for cp in self._cp_iter:
            if self._terminated.is_set():
                break
            # We iterate over portdb.porttrees, since it's common to
            # tweak this attribute in order to adjust repo selection.
            for mytree in portdb.porttrees:
                if self._terminated.is_set():
                    break
                repo_config = portdb.repositories.get_repo_for_location(mytree)
                if not repo_config.create_manifest:
                    if repo_config.name not in disabled_repos:
                        disabled_repos.add(repo_config.name)
                        portage.writemsg(
                            _(
                                ">>> Skipping creating Manifest for %s%s%s; "
                                "repository is configured to not use them\n"
                            )
                            % (cp, _repo_separator, repo_config.name),
                            noiselevel=-1,
                        )
                    continue
                cpv_list = portdb.cp_list(cp, mytree=[repo_config.location])
                if not cpv_list:
                    continue

                # Use _async_manifest_fetchlist(max_jobs=1), since we
                # spawn concurrent ManifestTask instances.
                yield ManifestTask(
                    cp=cp,
                    distdir=distdir,
                    fetchlist_dict=_async_manifest_fetchlist(
                        portdb,
                        repo_config,
                        cp,
                        cpv_list=cpv_list,
                        max_jobs=1,
                        loop=self._event_loop,
                    ),
                    repo_config=repo_config,
                    gpg_cmd=self._gpg_cmd,
                    gpg_vars=self._gpg_vars,
                    force_sign_key=self._force_sign_key,
                )

    def _task_exit(self, task):
        if task.returncode == os.EX_OK:
            self.cp_successful.add(task.cp)
        else:
            self.cp_failed.add(task.cp)
            if task.returncode != 1:
                # Retry if the returncode was unexpected.
                self.cp_retry.add(task.cp)

            if not self._terminated_tasks:
                portage.writemsg(
                    "Error processing %s%s%s with returncode %s, continuing...\n"
                    % (
                        task.cp,
                        _repo_separator,
                        task.repo_config.name,
                        task.returncode,
                    ),
                    noiselevel=-1,
                )

        AsyncScheduler._task_exit(self, task)



# (c) 2022-2025 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import contextlib
import json
import logging
import sys
import tempfile
import typing

from pathlib import Path, PurePath

from gpep517.build import build_wheel_impl
from gpep517.install import install_wheel_impl
from gpep517.utils import DEFAULT_PREFIX, get_toml, install_scheme_dict


ALL_OPT_LEVELS = [0, 1, 2]
DEFAULT_FALLBACK_BACKEND = "setuptools.build_meta:__legacy__"


def open_output(fd: int) -> typing.ContextManager[typing.TextIO]:
    """Safely return file open for outputting into fd"""
    if fd == 0:
        raise RuntimeError("--output-fd 0 invalid")
    elif fd == 1:
        return contextlib.nullcontext(sys.stdout)
    elif fd == 2:
        return contextlib.nullcontext(sys.stderr)
    return open(fd, "w")


def get_backend(args):
    with open_output(args.output_fd) as out:
        print(get_toml(args.pyproject_toml)
              .get("build-system", {})
              .get("build-backend", ""),
              file=out)
    return 0


def build_wheel(args):
    with open_output(args.output_fd) as out:
        print(build_wheel_impl(args, args.wheel_dir), file=out)
    return 0


def parse_optimize_arg(val):
    spl = val.split(",")
    if "all" in spl:
        return ALL_OPT_LEVELS
    return [int(x) for x in spl]


def install_wheel(args):
    install_wheel_impl(args, args.wheel)

    return 0


def install_from_source(args):
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        wheel = build_wheel_impl(args, temp_path)
        install_wheel_impl(args, temp_path / wheel)

    return 0


def verify_pyc(args):
    from gpep517.qa import qa_verify_pyc

    install_dict = install_scheme_dict(args.prefix or DEFAULT_PREFIX, "")
    sitedirs = frozenset(Path(install_dict[x]) for x in ("purelib", "platlib"))
    result = qa_verify_pyc(args.destdir, sitedirs)

    def fpath(p):
        if isinstance(p, Path):
            return str(p.root / p.relative_to(args.destdir))
        return p

    for kind, entries in result.items():
        for e in sorted(entries):
            print(f"{kind}:{':'.join(fpath(x) for x in e)}")
    return 1 if any(v for v in result.values()) else 0


def add_prefix_args(parser):
    parser.add_argument("--prefix",
                        type=Path,
                        help="Prefix to install to "
                        f"(default: {DEFAULT_PREFIX})")


def add_install_path_args(parser):
    parser.add_argument("--destdir",
                        type=Path,
                        help="Staging directory for the install (it will "
                        "be prepended to all paths)",
                        required=True)


def add_build_args(parser):
    group = parser.add_argument_group("backend selection")
    group.add_argument("--backend",
                       help="Backend to use (defaults to reading "
                            "from pyproject.toml)")
    group.add_argument("--fallback-backend",
                       default=DEFAULT_FALLBACK_BACKEND,
                       help="Backend to use if pyproject.toml does not exist "
                       "or does not specify one "
                       f"(default: {DEFAULT_FALLBACK_BACKEND!r})")
    group.add_argument("--no-fallback-backend",
                       action="store_const",
                       dest="fallback_backend",
                       const=None,
                       help="Disable backend fallback (i.e. require backend "
                       "declaration in pyproject.toml")
    group.add_argument("--pyproject-toml",
                       type=Path,
                       default="pyproject.toml",
                       help="Path to pyproject.toml file (used only if "
                       "--backend is not specified)")

    group = parser.add_argument_group("build options")
    group.add_argument("--allow-compressed",
                       help="Allow creating compressed zipfiles (gpep517 "
                       "will attempt to patch compression out by default)",
                       action="store_true")
    group.add_argument("--config-json",
                       help="JSON-encoded dictionary of config_settings "
                            "to pass to the build backend",
                       type=json.loads)
    group.add_argument("--sysroot",
                       help="Patch sysconfig paths to use specified sysroot "
                            "(experimental cross-compilation support)",
                       type=Path)


def add_install_args(parser):
    add_install_path_args(parser)

    group = parser.add_argument_group("install options")
    group.add_argument("--interpreter",
                       type=Path,
                       default=sys.executable,
                       help="The interpreter to put in script shebangs "
                       f"(default: {sys.executable})")
    group.add_argument("--optimize",
                       type=parse_optimize_arg,
                       default=[],
                       help="Comma-separated list of optimization levels "
                       "to compile bytecode for (default: none), pass 'all' "
                       "to enable all known optimization levels (currently: "
                       f"{', '.join(str(x) for x in ALL_OPT_LEVELS)})")
    group.add_argument("--overwrite",
                       action="store_true",
                       help="Permit overwriting files in destdir")
    group.add_argument("--symlink-pyc",
                       action="store_true",
                       help="Symlink .pyc files between optimization levels "
                       "if their contents match")

    group.add_argument("--symlink-to",
                       type=PurePath,
                       help="Install symlinks to another directory rather "
                       "than files if they match respective paths "
                       "in the other directory (useful for deduplicating "
                       "packages across Python implementations)")


def main(argv=sys.argv):
    argp = argparse.ArgumentParser(prog=argv[0])
    argp.add_argument("-q", "--quiet",
                      action="store_const",
                      dest="loglevel",
                      const=logging.WARNING,
                      default=logging.INFO,
                      help="Disable verbose progress reporting")

    subp = argp.add_subparsers(dest="command",
                               required=True)

    parser = subp.add_parser("get-backend",
                             help="Print build-backend from pyproject.toml")
    parser.add_argument("--output-fd",
                        default=1,
                        help="FD to use for output (default: 1)",
                        type=int)
    parser.add_argument("--pyproject-toml",
                        type=Path,
                        default="pyproject.toml",
                        help="Path to pyproject.toml file")

    parser = subp.add_parser("build-wheel",
                             help="Build wheel from sources")
    group = parser.add_argument_group("required arguments")
    group.add_argument("--output-fd",
                       help="FD to output the wheel name to",
                       required=True,
                       type=int)
    group.add_argument("--wheel-dir",
                       type=Path,
                       help="Directory to write the wheel into",
                       required=True)
    add_prefix_args(parser)
    add_build_args(parser)

    parser = subp.add_parser("install-from-source",
                             help="Build and install wheel from sources "
                             "(without preserving the wheel)")
    add_prefix_args(parser)
    add_build_args(parser)
    add_install_args(parser)

    parser = subp.add_parser("install-wheel",
                             help="Install the specified wheel")
    add_prefix_args(parser)
    add_install_args(parser)
    parser.add_argument("wheel",
                        type=Path,
                        help="Wheel to install")

    parser = subp.add_parser("verify-pyc",
                             help="Verify that all installed modules were "
                                  "byte-compiled and there are no stray .pyc "
                                  "files")
    add_prefix_args(parser)
    add_install_path_args(parser)

    args = argp.parse_args(argv[1:])
    logging.basicConfig(format="{asctime} {name} {levelname} {message}",
                        style="{",
                        level=args.loglevel)

    func = globals()[args.command.replace("-", "_")]
    return func(args)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2011-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.update import parse_updates
from portage.dep import Atom


class ParseUpdatesTestCase(TestCase):
    def testParseUpdates(self):
        test_cases = (
            (
                """
slotmove invalid_atom 0 3
slotmove !=invalid/blocker-3* 0 3
slotmove =valid/atom-3* 0 3 invalid_extra_token
slotmove =valid/atom-3* 0 3
slotmove =valid/atom-3* 0 3/3.1
slotmove =valid/atom-3* 0/0 3
move valid/atom1 valid/atom2 invalid_extra_token
move valid/atom1 invalid_atom2
move invalid_atom1 valid/atom2
move !invalid/blocker1 valid/atom2
move valid/atom1 !invalid/blocker2
move =invalid/operator-1* valid/atom2
move valid/atom1 =invalid/operator-2*
move valid/atom1 valid/atom2
""",
                [
                    ["slotmove", Atom("=valid/atom-3*"), "0", "3"],
                    ["move", Atom("valid/atom1"), Atom("valid/atom2")],
                ],
                12,
            ),
        )

        for input_content, expected_output, expected_error_count in test_cases:
            output_data, errors = parse_updates(input_content)
            self.assertEqual(output_data, expected_output)
            self.assertEqual(len(errors), expected_error_count)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2014-2021 The Meson development team

from __future__ import annotations

import os
import typing as T
import xml.etree.ElementTree as ET

from .vs2010backend import Vs2010Backend

if T.TYPE_CHECKING:
    from ..build import Build
    from ..interpreter import Interpreter


class Vs2022Backend(Vs2010Backend):

    name = 'vs2022'

    def __init__(self, build: T.Optional[Build], interpreter: T.Optional[Interpreter], gen_lite: bool = False):
        super().__init__(build, interpreter, gen_lite=gen_lite)
        self.sln_file_version = '12.00'
        self.sln_version_comment = 'Version 17'

    def detect_toolset(self) -> None:
        if self.environment is not None:
            comps = self.environment.coredata.compilers.host
            if comps and all(c.id == 'clang-cl' for c in comps.values()):
                self.platform_toolset = 'ClangCL'
            elif comps and all(c.id == 'intel-cl' for c in comps.values()):
                c = list(comps.values())[0]
                if c.version.startswith('19'):
                    self.platform_toolset = 'Intel C++ Compiler 19.0'
                # We don't have support for versions older than 2022 right now.
            if not self.platform_toolset:
                self.platform_toolset = 'v143'
            self.vs_version = '2022'
        # WindowsSDKVersion should be set by command prompt.
        sdk_version = os.environ.get('WindowsSDKVersion', None)
        if sdk_version:
            self.windows_target_platform_version = sdk_version.rstrip('\\')

    def generate_debug_information(self, link):
        # valid values for vs2022 is 'false', 'true', 'DebugFastLink', 'DebugFull'
        ET.SubElement(link, 'GenerateDebugInformation').text = 'DebugFull'

    def generate_lang_standard_info(self, file_args, clconf):
        if 'cpp' in file_args:
            optargs = [x for x in file_args['cpp'] if x.startswith('/std:c++')]
            if optargs:
                ET.SubElement(clconf, 'LanguageStandard').text = optargs[0].replace("/std:c++", "stdcpp")
        if 'c' in file_args:
            optargs = [x for x in file_args['c'] if x.startswith('/std:c')]
            if optargs:
                ET.SubElement(clconf, 'LanguageStandard_C').text = optargs[0].replace("/std:c", "stdc")



# Copyright 2012-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools

from portage import os
from portage.util._ctypes import load_libc
from portage.util._async.ForkProcess import ForkProcess


class SyncfsProcess(ForkProcess):
    """
    Isolate ctypes usage in a subprocess, in order to avoid
    potential problems with stale cached libraries as
    described in bug #448858, comment #14 (also see
    https://bugs.python.org/issue14597).
    """

    __slots__ = ("paths",)

    def _start(self):
        self.target = functools.partial(self._target, self._get_syncfs, self.paths)
        super()._start()

    @staticmethod
    def _get_syncfs():
        (libc, _) = load_libc()
        if libc is not None:
            return getattr(libc, "syncfs", None)
        return None

    @staticmethod
    def _target(get_syncfs, paths):
        syncfs_failed = False
        syncfs = get_syncfs()

        if syncfs is not None:
            for path in paths:
                try:
                    fd = os.open(path, os.O_RDONLY)
                except OSError:
                    pass
                else:
                    try:
                        if syncfs(fd) != 0:
                            # Happens with PyPy (bug #446610)
                            syncfs_failed = True
                    finally:
                        os.close(fd)

        if syncfs is None or syncfs_failed:
            return 1
        return os.EX_OK



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import logging

from _emerge.SpawnProcess import SpawnProcess
import portage
from portage.util.compression_probe import (
    compression_probe,
    _compressors,
)
from portage.util.cpuinfo import makeopts_to_job_count
from portage.process import find_binary
from portage.util import varexpand
from portage.exception import InvalidBinaryPackageFormat
from portage.binpkg import get_binpkg_format
import shlex
import signal
import subprocess
import tarfile
import textwrap


class BinpkgExtractorAsync(SpawnProcess):
    __slots__ = ("features", "image_dir", "pkg", "pkg_path")

    _shell_binary = portage.const.BASH_BINARY

    def _start(self):
        binpkg_format = get_binpkg_format(self.pkg_path)
        if binpkg_format == "xpak":
            self._xpak_start()
        else:
            raise InvalidBinaryPackageFormat(
                f"{self.pkg_path} is not a valid xpak binary package"
            )

    def _xpak_start(self):
        tar_options = ""
        if "xattr" in self.features:
            process = subprocess.Popen(
                ["gtar", "--help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            output = process.communicate()[0]
            if b"--xattrs" in output:
                tar_options = ["--xattrs", "--xattrs-include='*'"]
                for x in shlex.split(self.env.get("PORTAGE_XATTR_EXCLUDE", "")):
                    tar_options.append(shlex.quote(f"--xattrs-exclude={x}"))
                tar_options = " ".join(tar_options)

        decomp = _compressors.get(compression_probe(self.pkg_path))
        if decomp is not None:
            decomp_cmd = decomp.get("decompress")
            decomp_cmd = decomp_cmd.replace(
                "{JOBS}", str(makeopts_to_job_count(self.env.get("MAKEOPTS", "1")))
            )
        elif tarfile.is_tarfile(
            portage._unicode_encode(
                self.pkg_path, encoding=portage._encodings["fs"], errors="strict"
            )
        ):
            decomp_cmd = "cat"
            decomp = {
                "compress": "cat",
                "package": "sys-apps/coreutils",
            }
        else:
            decomp_cmd = None
        if decomp_cmd is None:
            self.scheduler.output(
                f"!!! File compression header unrecognized: {self.pkg_path}\n",
                log_path=self.logfile,
                background=self.background,
                level=logging.ERROR,
            )
            self.returncode = 1
            self._async_wait()
            return

        try:
            decompression_binary = shlex.split(varexpand(decomp_cmd, mydict=self.env))[
                0
            ]
        except IndexError:
            decompression_binary = ""

        if find_binary(decompression_binary) is None:
            # Try alternative command if it exists
            if decomp.get("decompress_alt"):
                decomp_cmd = decomp.get("decompress_alt")
            try:
                decompression_binary = shlex.split(
                    varexpand(decomp_cmd, mydict=self.env)
                )[0]
            except IndexError:
                decompression_binary = ""

            if find_binary(decompression_binary) is None:
                missing_package = decomp.get("package")
                self.scheduler.output(
                    f"!!! File compression unsupported {self.pkg_path}.\n"
                    f" Command was: {varexpand(decomp_cmd, mydict=self.env)}.\n"
                    f" Missing package: {missing_package}\n",
                    log_path=self.logfile,
                    background=self.background,
                    level=logging.ERROR,
                )
                self.returncode = 1
                self._async_wait()
                return

        pkg_xpak = portage.xpak.tbz2(self.pkg_path)
        pkg_xpak.scan()

        # SIGPIPE handling (128 + SIGPIPE) should be compatible with
        # assert_sigpipe_ok() that's used by the ebuild unpack() helper.
        self.args = [
            self._shell_binary,
            "-c",
            textwrap.dedent(
                f"""
                    cmd0=(head -c {pkg_xpak.filestat.st_size - pkg_xpak.xpaksize} -- {shlex.quote(self.pkg_path)})
                    cmd1=({decomp_cmd})
                    cmd2=(gtar -xp {tar_options} -C {shlex.quote(self.image_dir)} -f -);
                """
                """
                    "${cmd0[@]}" | "${cmd1[@]}" | "${cmd2[@]}";
                    p=(${PIPESTATUS[@]}) ; for i in {0..2}; do
                """
                f"""
                    if [[ ${{p[$i]}} != 0 && ${{p[$i]}} != {128 + signal.SIGPIPE} ]] ; then
                """
                """
                    echo command $(eval "echo \\"'\\${cmd$i[*]}'\\"") failed with status ${p[$i]} ;
                    exit ${p[$i]} ; fi ; done;
                    if [ ${p[$i]} != 0 ] ; then
                    echo command $(eval "echo \\"'\\${cmd$i[*]}'\\"") failed with status ${p[$i]} ;
                    exit ${p[$i]} ; fi ;
                    exit 0 ;
                """
            )
            .replace("\n", " ")
            .strip(),
        ]

        SpawnProcess._start(self)



# The most useful windows datatypes
import ctypes

BYTE = ctypes.c_ubyte
WORD = ctypes.c_ushort
DWORD = ctypes.c_ulong

#UCHAR = ctypes.c_uchar
CHAR = ctypes.c_char
WCHAR = ctypes.c_wchar
UINT = ctypes.c_uint
INT = ctypes.c_int

DOUBLE = ctypes.c_double
FLOAT = ctypes.c_float

BOOLEAN = BYTE
BOOL = ctypes.c_long

class VARIANT_BOOL(ctypes._SimpleCData):
    _type_ = "v"
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.value)

ULONG = ctypes.c_ulong
LONG = ctypes.c_long

USHORT = ctypes.c_ushort
SHORT = ctypes.c_short

# in the windows header files, these are structures.
_LARGE_INTEGER = LARGE_INTEGER = ctypes.c_longlong
_ULARGE_INTEGER = ULARGE_INTEGER = ctypes.c_ulonglong

LPCOLESTR = LPOLESTR = OLESTR = ctypes.c_wchar_p
LPCWSTR = LPWSTR = ctypes.c_wchar_p
LPCSTR = LPSTR = ctypes.c_char_p
LPCVOID = LPVOID = ctypes.c_void_p

# WPARAM is defined as UINT_PTR (unsigned type)
# LPARAM is defined as LONG_PTR (signed type)
if ctypes.sizeof(ctypes.c_long) == ctypes.sizeof(ctypes.c_void_p):
    WPARAM = ctypes.c_ulong
    LPARAM = ctypes.c_long
elif ctypes.sizeof(ctypes.c_longlong) == ctypes.sizeof(ctypes.c_void_p):
    WPARAM = ctypes.c_ulonglong
    LPARAM = ctypes.c_longlong

ATOM = WORD
LANGID = WORD

COLORREF = DWORD
LGRPID = DWORD
LCTYPE = DWORD

LCID = DWORD

################################################################
# HANDLE types
HANDLE = ctypes.c_void_p # in the header files: void *

HACCEL = HANDLE
HBITMAP = HANDLE
HBRUSH = HANDLE
HCOLORSPACE = HANDLE
HDC = HANDLE
HDESK = HANDLE
HDWP = HANDLE
HENHMETAFILE = HANDLE
HFONT = HANDLE
HGDIOBJ = HANDLE
HGLOBAL = HANDLE
HHOOK = HANDLE
HICON = HANDLE
HINSTANCE = HANDLE
HKEY = HANDLE
HKL = HANDLE
HLOCAL = HANDLE
HMENU = HANDLE
HMETAFILE = HANDLE
HMODULE = HANDLE
HMONITOR = HANDLE
HPALETTE = HANDLE
HPEN = HANDLE
HRGN = HANDLE
HRSRC = HANDLE
HSTR = HANDLE
HTASK = HANDLE
HWINSTA = HANDLE
HWND = HANDLE
SC_HANDLE = HANDLE
SERVICE_STATUS_HANDLE = HANDLE

################################################################
# Some important structure definitions

class RECT(ctypes.Structure):
    _fields_ = [("left", LONG),
                ("top", LONG),
                ("right", LONG),
                ("bottom", LONG)]
tagRECT = _RECTL = RECTL = RECT

class _SMALL_RECT(ctypes.Structure):
    _fields_ = [('Left', SHORT),
                ('Top', SHORT),
                ('Right', SHORT),
                ('Bottom', SHORT)]
SMALL_RECT = _SMALL_RECT

class _COORD(ctypes.Structure):
    _fields_ = [('X', SHORT),
                ('Y', SHORT)]

class POINT(ctypes.Structure):
    _fields_ = [("x", LONG),
                ("y", LONG)]
tagPOINT = _POINTL = POINTL = POINT

class SIZE(ctypes.Structure):
    _fields_ = [("cx", LONG),
                ("cy", LONG)]
tagSIZE = SIZEL = SIZE

def RGB(red, green, blue):
    return red + (green << 8) + (blue << 16)

class FILETIME(ctypes.Structure):
    _fields_ = [("dwLowDateTime", DWORD),
                ("dwHighDateTime", DWORD)]
_FILETIME = FILETIME

class MSG(ctypes.Structure):
    _fields_ = [("hWnd", HWND),
                ("message", UINT),
                ("wParam", WPARAM),
                ("lParam", LPARAM),
                ("time", DWORD),
                ("pt", POINT)]
tagMSG = MSG
MAX_PATH = 260

class WIN32_FIND_DATAA(ctypes.Structure):
    _fields_ = [("dwFileAttributes", DWORD),
                ("ftCreationTime", FILETIME),
                ("ftLastAccessTime", FILETIME),
                ("ftLastWriteTime", FILETIME),
                ("nFileSizeHigh", DWORD),
                ("nFileSizeLow", DWORD),
                ("dwReserved0", DWORD),
                ("dwReserved1", DWORD),
                ("cFileName", CHAR * MAX_PATH),
                ("cAlternateFileName", CHAR * 14)]

class WIN32_FIND_DATAW(ctypes.Structure):
    _fields_ = [("dwFileAttributes", DWORD),
                ("ftCreationTime", FILETIME),
                ("ftLastAccessTime", FILETIME),
                ("ftLastWriteTime", FILETIME),
                ("nFileSizeHigh", DWORD),
                ("nFileSizeLow", DWORD),
                ("dwReserved0", DWORD),
                ("dwReserved1", DWORD),
                ("cFileName", WCHAR * MAX_PATH),
                ("cAlternateFileName", WCHAR * 14)]

################################################################
# Pointer types

LPBOOL = PBOOL = ctypes.POINTER(BOOL)
PBOOLEAN = ctypes.POINTER(BOOLEAN)
LPBYTE = PBYTE = ctypes.POINTER(BYTE)
PCHAR = ctypes.POINTER(CHAR)
LPCOLORREF = ctypes.POINTER(COLORREF)
LPDWORD = PDWORD = ctypes.POINTER(DWORD)
LPFILETIME = PFILETIME = ctypes.POINTER(FILETIME)
PFLOAT = ctypes.POINTER(FLOAT)
LPHANDLE = PHANDLE = ctypes.POINTER(HANDLE)
PHKEY = ctypes.POINTER(HKEY)
LPHKL = ctypes.POINTER(HKL)
LPINT = PINT = ctypes.POINTER(INT)
PLARGE_INTEGER = ctypes.POINTER(LARGE_INTEGER)
PLCID = ctypes.POINTER(LCID)
LPLONG = PLONG = ctypes.POINTER(LONG)
LPMSG = PMSG = ctypes.POINTER(MSG)
LPPOINT = PPOINT = ctypes.POINTER(POINT)
PPOINTL = ctypes.POINTER(POINTL)
LPRECT = PRECT = ctypes.POINTER(RECT)
LPRECTL = PRECTL = ctypes.POINTER(RECTL)
LPSC_HANDLE = ctypes.POINTER(SC_HANDLE)
PSHORT = ctypes.POINTER(SHORT)
LPSIZE = PSIZE = ctypes.POINTER(SIZE)
LPSIZEL = PSIZEL = ctypes.POINTER(SIZEL)
PSMALL_RECT = ctypes.POINTER(SMALL_RECT)
LPUINT = PUINT = ctypes.POINTER(UINT)
PULARGE_INTEGER = ctypes.POINTER(ULARGE_INTEGER)
PULONG = ctypes.POINTER(ULONG)
PUSHORT = ctypes.POINTER(USHORT)
PWCHAR = ctypes.POINTER(WCHAR)
LPWIN32_FIND_DATAA = PWIN32_FIND_DATAA = ctypes.POINTER(WIN32_FIND_DATAA)
LPWIN32_FIND_DATAW = PWIN32_FIND_DATAW = ctypes.POINTER(WIN32_FIND_DATAW)
LPWORD = PWORD = ctypes.POINTER(WORD)



# portage: Installation
# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from enum import Enum

TYPES = Enum(
    "InstallationType",
    [
        "SOURCE",  # Portage is not installed, but running from its source tree.
        "MODULE",  # Portage is installed solely as a Python module.
        "SYSTEM",  # Portage is fully installed to the system, possibly prefixed.
    ],
)

if "SYSTEM" == "MODULE":
    TYPE = TYPES.MODULE
elif "SYSTEM" == "SYSTEM":
    TYPE = TYPES.SYSTEM
else:
    TYPE = TYPES.SOURCE



# SPDX-License-Identifier: Apache-2.0
# Copyright 2019 The Meson development team

from __future__ import annotations

import argparse
import subprocess
from pathlib import Path
import tempfile
import os
import shutil
import sys

from .run_tool import run_with_buffered_output, run_clang_tool_on_sources
from ..environment import detect_clangtidy, detect_clangapply
import typing as T

async def run_clang_tidy(fname: Path, tidyexe: list, builddir: Path, fixesdir: T.Optional[Path]) -> int:
    args = []
    if fixesdir is not None:
        handle, name = tempfile.mkstemp(prefix=fname.name + '.', suffix='.yaml', dir=fixesdir)
        os.close(handle)
        args.extend(['-export-fixes', name])
    return await run_with_buffered_output(tidyexe + args + ['-quiet', '-p', str(builddir), str(fname)])

def run(args: T.List[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('--fix', action='store_true')
    parser.add_argument('--color', default='always')
    parser.add_argument('sourcedir')
    parser.add_argument('builddir')
    options = parser.parse_args(args)

    srcdir = Path(options.sourcedir)
    builddir = Path(options.builddir)

    tidyexe = detect_clangtidy()
    if not tidyexe:
        print(f'Could not execute clang-tidy "{" ".join(tidyexe)}"')
        return 1

    if options.color == 'always' or options.color == 'auto' and sys.stdout.isatty():
        tidyexe += ['--use-color']

    fixesdir: T.Optional[Path] = None
    if options.fix:
        applyexe = detect_clangapply()
        if not applyexe:
            print(f'Could not execute clang-apply-replacements "{" ".join(applyexe)}"')
            return 1

        fixesdir = builddir / 'meson-private' / 'clang-tidy-fix'
        if fixesdir.is_dir():
            shutil.rmtree(fixesdir)
        elif fixesdir.exists():
            fixesdir.unlink()
        fixesdir.mkdir(parents=True)

    tidyret = run_clang_tool_on_sources('clang-tidy', srcdir, builddir, run_clang_tidy, tidyexe, builddir, fixesdir)
    if fixesdir is not None:
        print('Applying fix-its...')
        applyret = subprocess.run(applyexe + ['-format', '-style=file', '-ignore-insert-conflict', fixesdir]).returncode

    if tidyret != 0:
        print('Errors encountered while running clang-tidy', file=sys.stderr)
        return tidyret
    if fixesdir is not None and applyret != 0:
        print('Errors encountered while running clang-apply-replacements', file=sys.stderr)
        return applyret
    return 0



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage.dep import Atom
from portage.exception import InvalidData
from portage.versions import _pkg_str


class IndexedVardb:
    """
    A vardbapi interface that sacrifices validation in order to
    improve performance. It takes advantage of vardbdbapi._aux_cache,
    which is backed by vdb_metadata.pickle. Since _aux_cache is
    not updated for every single merge/unmerge (see
    _aux_cache_threshold), the list of packages is obtained directly
    from the real vardbapi instance. If a package is missing from
    _aux_cache, then its metadata is obtained using the normal
    (validated) vardbapi.aux_get method.

    For performance reasons, the match method only supports package
    name and version constraints.
    """

    # Match returns unordered results.
    match_unordered = True

    _copy_attrs = ("cpv_exists", "_aux_cache_keys", "_cpv_sort_ascending")

    def __init__(self, vardb):
        self._vardb = vardb

        for k in self._copy_attrs:
            setattr(self, k, getattr(vardb, k))

        self._cp_map = None

    def cp_all(self, sort=True):
        """
        Returns an ordered iterator instead of a list, so that search
        results can be displayed incrementally.
        """
        if self._cp_map is not None:
            return iter(sorted(self._cp_map)) if sort else iter(self._cp_map)

        delta_data = self._vardb._cache_delta.loadRace()
        if delta_data is None:
            return self._iter_cp_all()

        self._vardb._cache_delta.applyDelta(delta_data)

        self._cp_map = cp_map = {}
        for cpv in self._vardb._aux_cache["packages"]:
            try:
                cpv = _pkg_str(cpv, db=self._vardb)
            except InvalidData:
                continue

            cp_list = cp_map.get(cpv.cp)
            if cp_list is None:
                cp_list = []
                cp_map[cpv.cp] = cp_list
            cp_list.append(cpv)

        return iter(sorted(self._cp_map)) if sort else iter(self._cp_map)

    def _iter_cp_all(self):
        self._cp_map = cp_map = {}
        previous_cp = None
        for cpv in self._vardb._iter_cpv_all(sort=True):
            cp = portage.cpv_getkey(cpv)
            if cp is not None:
                cp_list = cp_map.get(cp)
                if cp_list is None:
                    cp_list = []
                    cp_map[cp] = cp_list
                cp_list.append(cpv)
                if previous_cp is not None and previous_cp != cp:
                    yield previous_cp
                previous_cp = cp

        if previous_cp is not None:
            yield previous_cp

    def match(self, atom):
        """
        For performance reasons, only package name and version
        constraints are supported, and the returned list is
        unordered.
        """
        if not isinstance(atom, Atom):
            atom = Atom(atom)
        cp_list = self._cp_map.get(atom.cp)
        if cp_list is None:
            return []

        if atom == atom.cp:
            return cp_list[:]
        return portage.match_from_list(atom, cp_list)

    def aux_get(self, cpv, attrs, myrepo=None):
        pkg_data = self._vardb._aux_cache["packages"].get(cpv)
        if (
            not isinstance(pkg_data, tuple)
            or len(pkg_data) != 2
            or not isinstance(pkg_data[1], dict)
        ):
            pkg_data = None
        if pkg_data is None:
            # It may be missing from _aux_cache due to
            # _aux_cache_threshold.
            return self._vardb.aux_get(cpv, attrs)
        metadata = pkg_data[1]
        return [metadata.get(k, "") for k in attrs]



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


from portage import _encodings, _unicode_encode
from portage.exception import PortageException
from portage.tests import TestCase
from _emerge.DependencyArg import DependencyArg
from _emerge.UseFlagDisplay import UseFlagDisplay


class StringFormatTestCase(TestCase):
    """
    Test that string formatting works correctly in the current interpreter,
    which may be either python2 or python3.
    """

    unicode_strings = (
        "\u2018",
        "\u2019",
    )

    def testDependencyArg(self):
        self.assertEqual(_encodings["content"], "utf_8")

        for arg_unicode in self.unicode_strings:
            arg_bytes = _unicode_encode(arg_unicode, encoding=_encodings["content"])
            dependency_arg = DependencyArg(arg=arg_unicode)

            formatted_str = f"{dependency_arg}"
            self.assertEqual(formatted_str, arg_unicode)

            # Test the __str__ method which returns unicode in python3
            formatted_str = f"{dependency_arg}"
            self.assertEqual(formatted_str, arg_unicode)

    def testPortageException(self):
        self.assertEqual(_encodings["content"], "utf_8")

        for arg_unicode in self.unicode_strings:
            arg_bytes = _unicode_encode(arg_unicode, encoding=_encodings["content"])
            e = PortageException(arg_unicode)

            formatted_str = f"{e}"
            self.assertEqual(formatted_str, arg_unicode)

            # Test the __str__ method which returns unicode in python3
            formatted_str = f"{e}"
            self.assertEqual(formatted_str, arg_unicode)

    def testUseFlagDisplay(self):
        self.assertEqual(_encodings["content"], "utf_8")

        for enabled in (True, False):
            for forced in (True, False):
                for arg_unicode in self.unicode_strings:
                    e = UseFlagDisplay(arg_unicode, enabled, forced)

                    formatted_str = f"{e}"
                    self.assertEqual(isinstance(formatted_str, str), True)

                    # Test the __str__ method which returns unicode in python3
                    formatted_str = f"{e}"
                    self.assertEqual(isinstance(formatted_str, str), True)



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2021-2025 Intel Corporation
from __future__ import annotations

"""Abstraction for Cython language compilers."""

import typing as T

from .. import options
from ..mesonlib import EnvironmentException, version_compare
from .compilers import Compiler

if T.TYPE_CHECKING:
    from ..options import MutableKeyedOptionDictType
    from ..environment import Environment
    from ..build import BuildTarget


class CythonCompiler(Compiler):

    """Cython Compiler."""

    language = 'cython'
    id = 'cython'

    def needs_static_linker(self) -> bool:
        # We transpile into C, so we don't need any linker
        return False

    def get_always_args(self) -> T.List[str]:
        return ['--fast-fail']

    def get_werror_args(self) -> T.List[str]:
        return ['-Werror']

    def get_output_args(self, outputname: str) -> T.List[str]:
        return ['-o', outputname]

    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
        # Cython doesn't have optimization levels itself, the underlying
        # compiler might though
        return []

    def get_dependency_gen_args(self, outtarget: str, outfile: str) -> T.List[str]:
        if version_compare(self.version, '>=0.29.33'):
            return ['-M']
        return []

    def get_depfile_suffix(self) -> str:
        return 'dep'

    def sanity_check(self, work_dir: str, environment: 'Environment') -> None:
        code = 'print("hello world")'
        with self.cached_compile(code, environment.coredata) as p:
            if p.returncode != 0:
                raise EnvironmentException(f'Cython compiler {self.id!r} cannot compile programs')

    def get_pic_args(self) -> T.List[str]:
        # We can lie here, it's fine
        return []

    def compute_parameters_with_absolute_paths(self, parameter_list: T.List[str],
                                               build_dir: str) -> T.List[str]:
        new: T.List[str] = []
        for i in parameter_list:
            new.append(i)

        return new

    def get_options(self) -> 'MutableKeyedOptionDictType':
        opts = super().get_options()

        key = self.form_compileropt_key('version')
        opts[key] = options.UserComboOption(
            self.make_option_name(key),
            'Python version to target',
            '3',
            choices=['2', '3'])

        key = self.form_compileropt_key('language')
        opts[key] = options.UserComboOption(
            self.make_option_name(key),
            'Output C or C++ files',
            'c',
            choices=['c', 'cpp'])

        return opts

    def get_option_compile_args(self, target: 'BuildTarget', env: 'Environment', subproject: T.Optional[str] = None) -> T.List[str]:
        args: T.List[str] = []
        version = self.get_compileropt_value('version', env, target, subproject)
        assert isinstance(version, str)
        args.append(f'-{version}')

        lang = self.get_compileropt_value('language', env, target, subproject)
        assert isinstance(lang, str)
        if lang == 'cpp':
            args.append('--cplus')
        return args



# Copyright 2014-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictUnsatisfiedDeepDepsTestCase(TestCase):
    def testSlotConflictUnsatisfiedDeepDeps(self):
        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {"KEYWORDS": "~x86"},
            "dev-libs/B-1": {"DEPEND": "dev-libs/A"},
            "dev-libs/C-1": {"DEPEND": ">=dev-libs/A-2"},
            "dev-libs/D-1": {"DEPEND": "dev-libs/A"},
        }

        installed = {
            "dev-libs/broken-1": {
                "RDEPEND": "dev-libs/A dev-libs/initially-unsatisfied"
            },
        }

        world = (
            "dev-libs/A",
            "dev-libs/B",
            "dev-libs/C",
            "dev-libs/D",
            "dev-libs/broken",
        )

        test_cases = (
            # Test bug #520950, where unsatisfied deps of installed
            # packages are supposed to be ignored when they are beyond
            # the depth requested by the user.
            ResolverPlaygroundTestCase(
                ["dev-libs/B", "dev-libs/C", "dev-libs/D"],
                all_permutations=True,
                options={"--autounmask": "y", "--complete-graph": True},
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                unsatisfied_deps=[],
                success=False,
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--autounmask": "y", "--complete-graph": True},
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                unsatisfied_deps=["dev-libs/broken"],
                success=False,
            ),
            # Test --selective with --deep = 0
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--autounmask": "y",
                    "--complete-graph": True,
                    "--selective": True,
                    "--deep": 0,
                },
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                unsatisfied_deps=[],
                success=False,
            ),
            # Test --deep = 1
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--autounmask": "y",
                    "--autounmask-backtrack": "y",
                    "--complete-graph": True,
                    "--selective": True,
                    "--deep": 1,
                },
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                unsatisfied_deps=["dev-libs/initially-unsatisfied"],
                success=False,
            ),
            # With --autounmask-backtrack=y:
            # [ebuild  N    ~] dev-libs/A-2
            # [ebuild  N     ] dev-libs/C-1
            # [ebuild  N     ] dev-libs/D-1
            # [ebuild  N     ] dev-libs/B-1
            #
            # The following keyword changes are necessary to proceed:
            # (see "package.accept_keywords" in the portage(5) man page for more details)
            # required by dev-libs/C-1::test_repo
            # required by @selected
            # required by @world (argument)
            # =dev-libs/A-2 ~x86
            #
            #!!! Problems have been detected with your world file
            #!!! Please run emaint --check world
            #
            #
            #!!! Ebuilds for the following packages are either all
            #!!! masked or don't exist:
            # dev-libs/broken
            #
            # emerge: there are no ebuilds to satisfy "dev-libs/initially-unsatisfied".
            # (dependency required by "dev-libs/broken-1::test_repo" [installed])
            # (dependency required by "@selected" [set])
            # (dependency required by "@world" [argument])
            # Without --autounmask-backtrack=y:
            #!!! Multiple package instances within a single package slot have been pulled
            #!!! into the dependency graph, resulting in a slot conflict:
            #
            # dev-libs/A:0
            #
            #  (dev-libs/A-1:0/0::test_repo, ebuild scheduled for merge) pulled in by
            #    (no parents that aren't satisfied by other packages in this slot)
            #
            #  (dev-libs/A-2:0/0::test_repo, ebuild scheduled for merge) pulled in by
            #    >=dev-libs/A-2 required by (dev-libs/C-1:0/0::test_repo, ebuild scheduled for merge)
            #    ^^           ^
            #
            # The following keyword changes are necessary to proceed:
            # (see "package.accept_keywords" in the portage(5) man page for more details)
            # required by dev-libs/C-1::test_repo
            # required by @selected
            # required by @world (argument)
            # =dev-libs/A-2 ~x86
            #
            # emerge: there are no ebuilds to satisfy "dev-libs/initially-unsatisfied".
            # (dependency required by "dev-libs/broken-1::test_repo" [installed])
            # (dependency required by "@selected" [set])
            # (dependency required by "@world" [argument])
            # Test --deep = True
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--autounmask": "y",
                    "--autounmask-backtrack": "y",
                    "--complete-graph": True,
                    "--selective": True,
                    "--deep": True,
                },
                mergelist=[
                    "dev-libs/A-2",
                    "dev-libs/B-1",
                    "dev-libs/C-1",
                    "dev-libs/D-1",
                ],
                ignore_mergelist_order=True,
                unstable_keywords=["dev-libs/A-2"],
                unsatisfied_deps=["dev-libs/initially-unsatisfied"],
                success=False,
            ),
            # The effects of --autounmask-backtrack are the same as the previous test case.
            # Both test cases can randomly succeed with --autounmask-backtrack=n, when
            # "backtracking due to unsatisfied dep" randomly occurs before the autounmask
            # unstable keyword change. It would be possible to eliminate backtracking here
            # by recognizing that there are no alternatives to satisfy the dev-libs/broken
            # atom in the world file. Then the test cases will consistently succeed with
            # --autounmask-backtrack=n.
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 The Meson development team

from __future__ import annotations
import typing as T

from .backends import Backend
from .. import mlog
from ..mesonlib import MesonBugException


class NoneBackend(Backend):

    name = 'none'

    def generate(self, capture: bool = False, vslite_ctx: T.Optional[T.Dict] = None) -> None:
        # Check for (currently) unexpected capture arg use cases -
        if capture:
            raise MesonBugException('We do not expect the none backend to generate with \'capture = True\'')
        if vslite_ctx:
            raise MesonBugException('We do not expect the none backend to be given a valid \'vslite_ctx\'')

        if self.build.get_targets():
            raise MesonBugException('None backend cannot generate target rules, but should have failed earlier.')
        mlog.log('Generating simple install-only backend')
        self.serialize_tests()
        self.create_install_data_files()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2012-2017 The Meson development team
# Copyright © 2023-2025 Intel Corporation

from __future__ import annotations

import enum
import os.path
import string
import typing as T

from .. import options
from .. import mlog
from ..mesonlib import (
    EnvironmentException, Popen_safe,
    is_windows, LibType, version_compare
)
from .compilers import Compiler, CompileCheckMode

if T.TYPE_CHECKING:
    from ..build import BuildTarget
    from ..options import MutableKeyedOptionDictType
    from ..dependencies import Dependency
    from ..environment import Environment  # noqa: F401
    from ..envconfig import MachineInfo
    from ..linkers.linkers import DynamicLinker
    from ..mesonlib import MachineChoice


cuda_optimization_args: T.Dict[str, T.List[str]] = {
    'plain': [],
    '0': ['-G'],
    'g': ['-O0'],
    '1': ['-O1'],
    '2': ['-O2', '-lineinfo'],
    '3': ['-O3'],
    's': ['-O3']
}

cuda_debug_args: T.Dict[bool, T.List[str]] = {
    False: [],
    True: ['-g']
}


class Phase(enum.Enum):

    COMPILER = 'compiler'
    LINKER = 'linker'


class CudaCompiler(Compiler):

    LINKER_PREFIX = '-Xlinker='
    language = 'cuda'

    # NVCC flags taking no arguments.
    _FLAG_PASSTHRU_NOARGS = {
        # NVCC --long-option,                   NVCC -short-option              CUDA Toolkit 11.2.1 Reference
        '--objdir-as-tempdir',                  '-objtemp',                     # 4.2.1.2
        '--generate-dependency-targets',        '-MP',                          # 4.2.1.12
        '--allow-unsupported-compiler',         '-allow-unsupported-compiler',  # 4.2.1.14
        '--link',                                                               # 4.2.2.1
        '--lib',                                '-lib',                         # 4.2.2.2
        '--device-link',                        '-dlink',                       # 4.2.2.3
        '--device-c',                           '-dc',                          # 4.2.2.4
        '--device-w',                           '-dw',                          # 4.2.2.5
        '--cuda',                               '-cuda',                        # 4.2.2.6
        '--compile',                            '-c',                           # 4.2.2.7
        '--fatbin',                             '-fatbin',                      # 4.2.2.8
        '--cubin',                              '-cubin',                       # 4.2.2.9
        '--ptx',                                '-ptx',                         # 4.2.2.10
        '--preprocess',                         '-E',                           # 4.2.2.11
        '--generate-dependencies',              '-M',                           # 4.2.2.12
        '--generate-nonsystem-dependencies',    '-MM',                          # 4.2.2.13
        '--generate-dependencies-with-compile', '-MD',                          # 4.2.2.14
        '--generate-nonsystem-dependencies-with-compile', '-MMD',               # 4.2.2.15
        '--run',                                                                # 4.2.2.16
        '--profile',                            '-pg',                          # 4.2.3.1
        '--debug',                              '-g',                           # 4.2.3.2
        '--device-debug',                       '-G',                           # 4.2.3.3
        '--extensible-whole-program',           '-ewp',                         # 4.2.3.4
        '--generate-line-info',                 '-lineinfo',                    # 4.2.3.5
        '--dlink-time-opt',                     '-dlto',                        # 4.2.3.8
        '--no-exceptions',                      '-noeh',                        # 4.2.3.11
        '--shared',                             '-shared',                      # 4.2.3.12
        '--no-host-device-initializer-list',    '-nohdinitlist',                # 4.2.3.15
        '--expt-relaxed-constexpr',             '-expt-relaxed-constexpr',      # 4.2.3.16
        '--extended-lambda',                    '-extended-lambda',             # 4.2.3.17
        '--expt-extended-lambda',               '-expt-extended-lambda',        # 4.2.3.18
        '--m32',                                '-m32',                         # 4.2.3.20
        '--m64',                                '-m64',                         # 4.2.3.21
        '--forward-unknown-to-host-compiler',   '-forward-unknown-to-host-compiler', # 4.2.5.1
        '--forward-unknown-to-host-linker',     '-forward-unknown-to-host-linker',   # 4.2.5.2
        '--dont-use-profile',                   '-noprof',                      # 4.2.5.3
        '--dryrun',                             '-dryrun',                      # 4.2.5.5
        '--verbose',                            '-v',                           # 4.2.5.6
        '--keep',                               '-keep',                        # 4.2.5.7
        '--save-temps',                         '-save-temps',                  # 4.2.5.9
        '--clean-targets',                      '-clean',                       # 4.2.5.10
        '--no-align-double',                                                    # 4.2.5.16
        '--no-device-link',                     '-nodlink',                     # 4.2.5.17
        '--allow-unsupported-compiler',         '-allow-unsupported-compiler',  # 4.2.5.18
        '--use_fast_math',                      '-use_fast_math',               # 4.2.7.7
        '--extra-device-vectorization',         '-extra-device-vectorization',  # 4.2.7.12
        '--compile-as-tools-patch',             '-astoolspatch',                # 4.2.7.13
        '--keep-device-functions',              '-keep-device-functions',       # 4.2.7.14
        '--disable-warnings',                   '-w',                           # 4.2.8.1
        '--source-in-ptx',                      '-src-in-ptx',                  # 4.2.8.2
        '--restrict',                           '-restrict',                    # 4.2.8.3
        '--Wno-deprecated-gpu-targets',         '-Wno-deprecated-gpu-targets',  # 4.2.8.4
        '--Wno-deprecated-declarations',        '-Wno-deprecated-declarations', # 4.2.8.5
        '--Wreorder',                           '-Wreorder',                    # 4.2.8.6
        '--Wdefault-stream-launch',             '-Wdefault-stream-launch',      # 4.2.8.7
        '--Wext-lambda-captures-this',          '-Wext-lambda-captures-this',   # 4.2.8.8
        '--display-error-number',               '-err-no',                      # 4.2.8.10
        '--resource-usage',                     '-res-usage',                   # 4.2.8.14
        '--help',                               '-h',                           # 4.2.8.15
        '--version',                            '-V',                           # 4.2.8.16
        '--list-gpu-code',                      '-code-ls',                     # 4.2.8.20
        '--list-gpu-arch',                      '-arch-ls',                     # 4.2.8.21
    }
    # Dictionary of NVCC flags taking either one argument or a comma-separated list.
    # Maps --long to -short options, because the short options are more GCC-like.
    _FLAG_LONG2SHORT_WITHARGS = {
        '--output-file':                        '-o',                           # 4.2.1.1
        '--pre-include':                        '-include',                     # 4.2.1.3
        '--library':                            '-l',                           # 4.2.1.4
        '--define-macro':                       '-D',                           # 4.2.1.5
        '--undefine-macro':                     '-U',                           # 4.2.1.6
        '--include-path':                       '-I',                           # 4.2.1.7
        '--system-include':                     '-isystem',                     # 4.2.1.8
        '--library-path':                       '-L',                           # 4.2.1.9
        '--output-directory':                   '-odir',                        # 4.2.1.10
        '--dependency-output':                  '-MF',                          # 4.2.1.11
        '--compiler-bindir':                    '-ccbin',                       # 4.2.1.13
        '--archiver-binary':                    '-arbin',                       # 4.2.1.15
        '--cudart':                             '-cudart',                      # 4.2.1.16
        '--cudadevrt':                          '-cudadevrt',                   # 4.2.1.17
        '--libdevice-directory':                '-ldir',                        # 4.2.1.18
        '--target-directory':                   '-target-dir',                  # 4.2.1.19
        '--optimization-info':                  '-opt-info',                    # 4.2.3.6
        '--optimize':                           '-O',                           # 4.2.3.7
        '--ftemplate-backtrace-limit':          '-ftemplate-backtrace-limit',   # 4.2.3.9
        '--ftemplate-depth':                    '-ftemplate-depth',             # 4.2.3.10
        '--x':                                  '-x',                           # 4.2.3.13
        '--std':                                '-std',                         # 4.2.3.14
        '--machine':                            '-m',                           # 4.2.3.19
        '--compiler-options':                   '-Xcompiler',                   # 4.2.4.1
        '--linker-options':                     '-Xlinker',                     # 4.2.4.2
        '--archive-options':                    '-Xarchive',                    # 4.2.4.3
        '--ptxas-options':                      '-Xptxas',                      # 4.2.4.4
        '--nvlink-options':                     '-Xnvlink',                     # 4.2.4.5
        '--threads':                            '-t',                           # 4.2.5.4
        '--keep-dir':                           '-keep-dir',                    # 4.2.5.8
        '--run-args':                           '-run-args',                    # 4.2.5.11
        '--input-drive-prefix':                 '-idp',                         # 4.2.5.12
        '--dependency-drive-prefix':            '-ddp',                         # 4.2.5.13
        '--drive-prefix':                       '-dp',                          # 4.2.5.14
        '--dependency-target-name':             '-MT',                          # 4.2.5.15
        '--default-stream':                     '-default-stream',              # 4.2.6.1
        '--gpu-architecture':                   '-arch',                        # 4.2.7.1
        '--gpu-code':                           '-code',                        # 4.2.7.2
        '--generate-code':                      '-gencode',                     # 4.2.7.3
        '--relocatable-device-code':            '-rdc',                         # 4.2.7.4
        '--entries':                            '-e',                           # 4.2.7.5
        '--maxrregcount':                       '-maxrregcount',                # 4.2.7.6
        '--ftz':                                '-ftz',                         # 4.2.7.8
        '--prec-div':                           '-prec-div',                    # 4.2.7.9
        '--prec-sqrt':                          '-prec-sqrt',                   # 4.2.7.10
        '--fmad':                               '-fmad',                        # 4.2.7.11
        '--Werror':                             '-Werror',                      # 4.2.8.9
        '--diag-error':                         '-diag-error',                  # 4.2.8.11
        '--diag-suppress':                      '-diag-suppress',               # 4.2.8.12
        '--diag-warn':                          '-diag-warn',                   # 4.2.8.13
        '--options-file':                       '-optf',                        # 4.2.8.17
        '--time':                               '-time',                        # 4.2.8.18
        '--qpp-config':                         '-qpp-config',                  # 4.2.8.19
    }
    # Reverse map -short to --long options.
    _FLAG_SHORT2LONG_WITHARGS = {v: k for k, v in _FLAG_LONG2SHORT_WITHARGS.items()}

    id = 'nvcc'

    def __init__(self, ccache: T.List[str], exelist: T.List[str], version: str, for_machine: MachineChoice,
                 is_cross: bool,
                 host_compiler: Compiler, info: 'MachineInfo',
                 linker: T.Optional['DynamicLinker'] = None,
                 full_version: T.Optional[str] = None):
        super().__init__(ccache, exelist, version, for_machine, info, linker=linker, full_version=full_version, is_cross=is_cross)
        self.host_compiler = host_compiler
        self.base_options = host_compiler.base_options
        # -Wpedantic generates useless churn due to nvcc's dual compilation model producing
        # a temporary host C++ file that includes gcc-style line directives:
        # https://stackoverflow.com/a/31001220
        self.warn_args = {
            level: self._to_host_flags(list(f for f in flags if f != '-Wpedantic'))
            for level, flags in host_compiler.warn_args.items()
        }
        self.host_werror_args = ['-Xcompiler=' + x for x in self.host_compiler.get_werror_args()]
        self.debug_macros_available = version_compare(self.version, '>=12.9')

    @classmethod
    def _shield_nvcc_list_arg(cls, arg: str, listmode: bool = True) -> str:
        r"""
        Shield an argument against both splitting by NVCC's list-argument
        parse logic, and interpretation by any shell.

        NVCC seems to consider every comma , that is neither escaped by \ nor inside
        a double-quoted string a split-point. Single-quotes do not provide protection
        against splitting; In fact, after splitting they are \-escaped. Unfortunately,
        double-quotes don't protect against shell expansion. What follows is a
        complex dance to accommodate everybody.
        """

        SQ = "'"
        DQ = '"'
        CM = ","
        BS = "\\"
        DQSQ = DQ+SQ+DQ
        quotable = set(string.whitespace+'"$`\\')

        if CM not in arg or not listmode:
            if SQ not in arg:
                # If any of the special characters "$`\ or whitespace are present, single-quote.
                # Otherwise return bare.
                if set(arg).intersection(quotable):
                    return SQ+arg+SQ
                else:
                    return arg # Easy case: no splits, no quoting.
            else:
                # There are single quotes. Double-quote them, and single-quote the
                # strings between them.
                l = [cls._shield_nvcc_list_arg(s) for s in arg.split(SQ)]
                l = sum([[s, DQSQ] for s in l][:-1], [])  # Interleave l with DQSQs
                return ''.join(l)
        else:
            # A comma is present, and list mode was active.
            # We apply (what we guess is) the (primitive) NVCC splitting rule:
            l = ['']
            instring = False
            argit = iter(arg)
            for c in argit:
                if c == CM and not instring:
                    l.append('')
                elif c == DQ:
                    l[-1] += c
                    instring = not instring
                elif c == BS:
                    try:
                        l[-1] += next(argit)
                    except StopIteration:
                        break
                else:
                    l[-1] += c

            # Shield individual strings, without listmode, then return them with
            # escaped commas between them.
            l = [cls._shield_nvcc_list_arg(s, listmode=False) for s in l]
            return r'\,'.join(l)

    @classmethod
    def _merge_flags(cls, flags: T.List[str]) -> T.List[str]:
        r"""
        The flags to NVCC gets exceedingly verbose and unreadable when too many of them
        are shielded with -Xcompiler. Merge consecutive -Xcompiler-wrapped arguments
        into one.
        """
        if len(flags) <= 1:
            return flags
        flagit = iter(flags)
        xflags = []

        def is_xcompiler_flag_isolated(flag: str) -> bool:
            return flag == '-Xcompiler'

        def is_xcompiler_flag_glued(flag: str) -> bool:
            return flag.startswith('-Xcompiler=')

        def is_xcompiler_flag(flag: str) -> bool:
            return is_xcompiler_flag_isolated(flag) or is_xcompiler_flag_glued(flag)

        def get_xcompiler_val(flag: str, flagit: T.Iterator[str]) -> str:
            if is_xcompiler_flag_glued(flag):
                return flag[len('-Xcompiler='):]
            else:
                try:
                    return next(flagit)
                except StopIteration:
                    return ""

        ingroup = False
        for flag in flagit:
            if not is_xcompiler_flag(flag):
                ingroup = False
                xflags.append(flag)
            elif ingroup:
                xflags[-1] += ','
                xflags[-1] += get_xcompiler_val(flag, flagit)
            elif is_xcompiler_flag_isolated(flag):
                ingroup = True
                xflags.append(flag)
                xflags.append(get_xcompiler_val(flag, flagit))
            elif is_xcompiler_flag_glued(flag):
                ingroup = True
                xflags.append(flag)
            else:
                raise ValueError("-Xcompiler flag merging failed, unknown argument form!")
        return xflags

    @classmethod
    def to_host_flags_base(cls, flags: T.List[str], phase: Phase = Phase.COMPILER, default_include_dirs: T.Optional[T.List[str]] = None) -> T.List[str]:
        """
        Translate generic "GCC-speak" plus particular "NVCC-speak" flags to NVCC flags.

        NVCC's "short" flags have broad similarities to the GCC standard, but have
        gratuitous, irritating differences.
        """
        xflags = []
        flagit = iter(flags)

        for flag in flagit:
            # The CUDA Toolkit Documentation, in 4.1. Command Option Types and Notation,
            # specifies that NVCC does not parse the standard flags as GCC does. It has
            # its own strategy, to wit:
            #
            #     nvcc recognizes three types of command options: boolean options, single
            #     value options, and list options.
            #
            #     Boolean options do not have an argument; they are either specified on a
            #     command line or not. Single value options must be specified at most once,
            #     and list options may be repeated. Examples of each of these option types
            #     are, respectively: --verbose (switch to verbose mode), --output-file
            #     (specify output file), and --include-path (specify include path).
            #
            #     Single value options and list options must have arguments, which must
            #     follow the name of the option itself by either one of more spaces or an
            #     equals character. When a one-character short name such as -I, -l, and -L
            #     is used, the value of the option may also immediately follow the option
            #     itself without being separated by spaces or an equal character. The
            #     individual values of list options may be separated by commas in a single
            #     instance of the option, or the option may be repeated, or any
            #     combination of these two cases.
            #
            # One strange consequence of this choice is that directory and filenames that
            # contain commas (',') cannot be passed to NVCC (at least, not as easily as
            # in GCC). Another strange consequence is that it is legal to supply flags
            # such as
            #
            #     -lpthread,rt,dl,util
            #     -l pthread,rt,dl,util
            #     -l=pthread,rt,dl,util
            #
            # and each of the above alternatives is equivalent to GCC-speak
            #
            #     -lpthread -lrt -ldl -lutil
            #     -l pthread -l rt -l dl -l util
            #     -l=pthread -l=rt -l=dl -l=util
            #
            # *With the exception of commas in the name*, GCC-speak for these list flags
            # is a strict subset of NVCC-speak, so we passthrough those flags.
            #
            # The -D macro-define flag is documented as somehow shielding commas from
            # splitting a definition. Balanced parentheses, braces and single-quotes
            # around the comma are not sufficient, but balanced double-quotes are. The
            # shielding appears to work with -l, -I, -L flags as well, for instance.
            #
            # Since our goal is to replicate GCC-speak as much as possible, we check for
            # commas in all list-arguments and shield them with double-quotes. We make
            # an exception for -D (where this would be value-changing) and -U (because
            # it isn't possible to define a macro with a comma in the name).

            if flag in cls._FLAG_PASSTHRU_NOARGS:
                xflags.append(flag)
                continue

            # Handle breakup of flag-values into a flag-part and value-part.
            if flag[:1] not in '-/':
                # This is not a flag. It's probably a file input. Pass it through.
                xflags.append(flag)
                continue
            elif flag[:1] == '/':
                # This is ambiguously either an MVSC-style /switch or an absolute path
                # to a file. For some magical reason the following works acceptably in
                # both cases.
                # We only want to prefix arguments that are NOT static archives, since
                # the latter could contain relocatable device code (-dc/-rdc=true).
                prefix = '' if flag.endswith('.a') else f'-X{phase.value}='
                wrap = '"' if ',' in flag else ''
                xflags.append(f'{prefix}{wrap}{flag}{wrap}')
                continue
            elif len(flag) >= 2 and flag[0] == '-' and flag[1] in 'IDULlmOxmte':
                # This is a single-letter short option. These options (with the
                # exception of -o) are allowed to receive their argument with neither
                # space nor = sign before them. Detect and separate them in that event.
                if flag[2:3] == '':            # -I something
                    try:
                        val = next(flagit)
                    except StopIteration:
                        pass
                elif flag[2:3] == '=':           # -I=something
                    val = flag[3:]
                else:                            # -Isomething
                    val = flag[2:]
                flag = flag[:2]                  # -I
            elif flag in cls._FLAG_LONG2SHORT_WITHARGS or \
                    flag in cls._FLAG_SHORT2LONG_WITHARGS:
                # This is either -o or a multi-letter flag, and it is receiving its
                # value isolated.
                try:
                    val = next(flagit)           # -o something
                except StopIteration:
                    pass
            elif flag.split('=', 1)[0] in cls._FLAG_LONG2SHORT_WITHARGS or \
                    flag.split('=', 1)[0] in cls._FLAG_SHORT2LONG_WITHARGS:
                # This is either -o or a multi-letter flag, and it is receiving its
                # value after an = sign.
                flag, val = flag.split('=', 1)    # -o=something
            # Some dependencies (e.g., BoostDependency) add unspaced "-isystem/usr/include" arguments
            elif flag.startswith('-isystem'):
                val = flag[8:].strip()
                flag = flag[:8]
            else:
                # This is a flag, and it's foreign to NVCC.
                #
                # We do not know whether this GCC-speak flag takes an isolated
                # argument. Assuming it does not (the vast majority indeed don't),
                # wrap this argument in an -Xcompiler flag and send it down to NVCC.
                if flag == '-ffast-math':
                    xflags.append('-use_fast_math')
                    xflags.append('-Xcompiler='+flag)
                elif flag == '-fno-fast-math':
                    xflags.append('-ftz=false')
                    xflags.append('-prec-div=true')
                    xflags.append('-prec-sqrt=true')
                    xflags.append('-Xcompiler='+flag)
                elif flag == '-freciprocal-math':
                    xflags.append('-prec-div=false')
                    xflags.append('-Xcompiler='+flag)
                elif flag == '-fno-reciprocal-math':
                    xflags.append('-prec-div=true')
                    xflags.append('-Xcompiler='+flag)
                else:
                    xflags.append('-Xcompiler='+cls._shield_nvcc_list_arg(flag))
                    # The above should securely handle GCC's -Wl, -Wa, -Wp, arguments.
                continue

            assert val is not None  # Should only trip if there is a missing argument.

            # Take care of the various NVCC-supported flags that need special handling.
            flag = cls._FLAG_LONG2SHORT_WITHARGS.get(flag, flag)

            if flag in {'-include', '-isystem', '-I', '-L', '-l'}:
                # These flags are known to GCC, but list-valued in NVCC. They potentially
                # require double-quoting to prevent NVCC interpreting the flags as lists
                # when GCC would not have done so.
                #
                # We avoid doing this quoting for -D to avoid redefining macros and for
                # -U because it isn't possible to define a macro with a comma in the name.
                # -U with comma arguments is impossible in GCC-speak (and thus unambiguous
                #in NVCC-speak, albeit unportable).
                if len(flag) == 2:
                    xflags.append(flag+cls._shield_nvcc_list_arg(val))
                elif flag == '-isystem' and default_include_dirs is not None and val in default_include_dirs:
                    # like GnuLikeCompiler, we have to filter out include directories specified
                    # with -isystem that overlap with the host compiler's search path
                    pass
                else:
                    xflags.append(flag)
                    xflags.append(cls._shield_nvcc_list_arg(val))
            elif flag == '-O':
                # Handle optimization levels GCC knows about that NVCC does not.
                if val == 'fast':
                    xflags.append('-O3')
                    xflags.append('-use_fast_math')
                    xflags.append('-Xcompiler')
                    xflags.append(flag+val)
                elif val in {'s', 'g', 'z'}:
                    xflags.append('-Xcompiler')
                    xflags.append(flag+val)
                else:
                    xflags.append(flag+val)
            elif flag in {'-D', '-U', '-m', '-t'}:
                xflags.append(flag+val)       # For style, keep glued.
            elif flag in {'-std'}:
                xflags.append(flag+'='+val)   # For style, keep glued.
            else:
                xflags.append(flag)
                xflags.append(val)

        return cls._merge_flags(xflags)

    def _to_host_flags(self, flags: T.List[str], phase: Phase = Phase.COMPILER) -> T.List[str]:
        return self.to_host_flags_base(flags, phase, self.host_compiler.get_default_include_dirs())

    def needs_static_linker(self) -> bool:
        return False

    def thread_link_flags(self, environment: 'Environment') -> T.List[str]:
        return self._to_host_flags(self.host_compiler.thread_link_flags(environment), Phase.LINKER)

    def sanity_check(self, work_dir: str, env: 'Environment') -> None:
        mlog.debug('Sanity testing ' + self.get_display_language() + ' compiler:', ' '.join(self.exelist))
        mlog.debug('Is cross compiler: %s.' % str(self.is_cross))

        sname = 'sanitycheckcuda.cu'
        code = r'''
        #include <cuda_runtime.h>
        #include <stdio.h>

        __global__ void kernel (void) {}

        int main(void){
            struct cudaDeviceProp prop;
            int count, i;
            cudaError_t ret = cudaGetDeviceCount(&count);
            if(ret != cudaSuccess){
                fprintf(stderr, "%d\n", (int)ret);
            }else{
                for(i=0;i<count;i++){
                    if(cudaGetDeviceProperties(&prop, i) == cudaSuccess){
                        fprintf(stdout, "%d.%d\n", prop.major, prop.minor);
                    }
                }
            }
            fflush(stderr);
            fflush(stdout);
            return 0;
        }
        '''
        binname = sname.rsplit('.', 1)[0]
        binname += '_cross' if self.is_cross else ''
        source_name = os.path.join(work_dir, sname)
        binary_name = os.path.join(work_dir, binname + '.exe')
        with open(source_name, 'w', encoding='utf-8') as ofile:
            ofile.write(code)

        # The Sanity Test for CUDA language will serve as both a sanity test
        # and a native-build GPU architecture detection test, useful later.
        #
        # For this second purpose, NVCC has very handy flags, --run and
        # --run-args, that allow one to run an application with the
        # environment set up properly. Of course, this only works for native
        # builds; For cross builds we must still use the exe_wrapper (if any).
        self.detected_cc = ''
        flags = []

        # Disable warnings, compile with statically-linked runtime for minimum
        # reliance on the system.
        flags += ['-w', '-cudart', 'static', source_name]

        # Use the -ccbin option, if available, even during sanity checking.
        # Otherwise, on systems where CUDA does not support the default compiler,
        # NVCC becomes unusable.
        flags += self.get_ccbin_args(None, env, '')

        # If cross-compiling, we can't run the sanity check, only compile it.
        if self.is_cross and not env.has_exe_wrapper():
            # Linking cross built apps is painful. You can't really
            # tell if you should use -nostdlib or not and for example
            # on OSX the compiler binary is the same but you need
            # a ton of compiler flags to differentiate between
            # arm and x86_64. So just compile.
            flags += self.get_compile_only_args()
        flags += self.get_output_args(binary_name)

        # Compile sanity check
        cmdlist = self.exelist + flags
        mlog.debug('Sanity check compiler command line: ', ' '.join(cmdlist))
        pc, stdo, stde = Popen_safe(cmdlist, cwd=work_dir)
        mlog.debug('Sanity check compile stdout: ')
        mlog.debug(stdo)
        mlog.debug('-----\nSanity check compile stderr:')
        mlog.debug(stde)
        mlog.debug('-----')
        if pc.returncode != 0:
            raise EnvironmentException(f'Compiler {self.name_string()} cannot compile programs.')

        # Run sanity check (if possible)
        if self.is_cross:
            return

        cmdlist = self.exelist + ['--run', f'"{binary_name}"']
        try:
            stdo, stde = self.run_sanity_check(env, cmdlist, work_dir)
        except EnvironmentException:
            raise EnvironmentException(f'Executables created by {self.language} compiler {self.name_string()} are not runnable.')

        # Interpret the result of the sanity test.
        # As mentioned above, it is not only a sanity test but also a GPU
        # architecture detection test.
        if stde == '':
            self.detected_cc = stdo

    def has_header_symbol(self, hname: str, symbol: str, prefix: str,
                          env: 'Environment', *,
                          extra_args: T.Union[None, T.List[str], T.Callable[[CompileCheckMode], T.List[str]]] = None,
                          dependencies: T.Optional[T.List['Dependency']] = None) -> T.Tuple[bool, bool]:
        if extra_args is None:
            extra_args = []
        fargs = {'prefix': prefix, 'header': hname, 'symbol': symbol}
        # Check if it's a C-like symbol
        t = '''{prefix}
        #include <{header}>
        int main(void) {{
            /* If it's not defined as a macro, try to use as a symbol */
            #ifndef {symbol}
                {symbol};
            #endif
            return 0;
        }}'''
        found, cached = self.compiles(t.format_map(fargs), env, extra_args=extra_args, dependencies=dependencies)
        if found:
            return True, cached
        # Check if it's a class or a template
        t = '''{prefix}
        #include <{header}>
        using {symbol};
        int main(void) {{
            return 0;
        }}'''
        return self.compiles(t.format_map(fargs), env, extra_args=extra_args, dependencies=dependencies)

    _CPP14_VERSION = '>=9.0'
    _CPP17_VERSION = '>=11.0'
    _CPP20_VERSION = '>=12.0'

    def get_options(self) -> 'MutableKeyedOptionDictType':
        cpp_stds = ['none', 'c++03', 'c++11']
        if version_compare(self.version, self._CPP14_VERSION):
            cpp_stds += ['c++14']
        if version_compare(self.version, self._CPP17_VERSION):
            cpp_stds += ['c++17']
        if version_compare(self.version, self._CPP20_VERSION):
            cpp_stds += ['c++20']

        opts = super().get_options()

        key = self.form_compileropt_key('std')
        opts[key] = options.UserComboOption(
            self.make_option_name(key),
            'C++ language standard to use with CUDA',
            'none',
            choices=cpp_stds)

        key = self.form_compileropt_key('ccbindir')
        opts[key] = options.UserStringOption(
            self.make_option_name(key),
            'CUDA non-default toolchain directory to use (-ccbin)',
            '')

        return opts

    def get_option_compile_args(self, target: 'BuildTarget', env: 'Environment', subproject: T.Optional[str] = None) -> T.List[str]:
        args = self.get_ccbin_args(target, env, subproject)

        try:
            host_compiler_args = self.host_compiler.get_option_compile_args(target, env, subproject)
        except KeyError:
            host_compiler_args = []
        return args + self._to_host_flags(host_compiler_args)

    def get_option_std_args(self, target: BuildTarget, env: Environment, subproject: T.Optional[str] = None) -> T.List[str]:
        # On Windows, the version of the C++ standard used by nvcc is dictated by
        # the combination of CUDA version and MSVC version; the --std= is thus ignored
        # and attempting to use it will result in a warning: https://stackoverflow.com/a/51272091/741027
        if not is_windows():
            std = self.get_compileropt_value('std', env, target, subproject)
            assert isinstance(std, str)
            if std != 'none':
                return ['--std=' + std]

        try:
            host_compiler_args = self.host_compiler.get_option_std_args(target, env, subproject)
        except KeyError:
            host_compiler_args = []
        return self._to_host_flags(host_compiler_args)

    def get_option_link_args(self, target: 'BuildTarget', env: 'Environment', subproject: T.Optional[str] = None) -> T.List[str]:
        args = self.get_ccbin_args(target, env, subproject)
        return args + self._to_host_flags(self.host_compiler.get_option_link_args(target, env, subproject), Phase.LINKER)

    def get_soname_args(self, env: 'Environment', prefix: str, shlib_name: str,
                        suffix: str, soversion: str,
                        darwin_versions: T.Tuple[str, str]) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_soname_args(
            env, prefix, shlib_name, suffix, soversion, darwin_versions), Phase.LINKER)

    def get_compile_only_args(self) -> T.List[str]:
        return ['-c']

    def get_no_optimization_args(self) -> T.List[str]:
        return ['-O0']

    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
        # alternatively, consider simply redirecting this to the host compiler, which would
        # give us more control over options like "optimize for space" (which nvcc doesn't support):
        # return self._to_host_flags(self.host_compiler.get_optimization_args(optimization_level))
        return cuda_optimization_args[optimization_level]

    def sanitizer_compile_args(self, value: T.List[str]) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.sanitizer_compile_args(value))

    def sanitizer_link_args(self, value: T.List[str]) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.sanitizer_link_args(value))

    def get_debug_args(self, is_debug: bool) -> T.List[str]:
        return cuda_debug_args[is_debug]

    def get_werror_args(self) -> T.List[str]:
        device_werror_args = ['-Werror=cross-execution-space-call,deprecated-declarations,reorder']
        return device_werror_args + self.host_werror_args

    def get_warn_args(self, level: str) -> T.List[str]:
        return self.warn_args[level]

    def get_include_args(self, path: str, is_system: bool) -> T.List[str]:
        if path == '':
            path = '.'
        return ['-isystem=' + path] if is_system else ['-I' + path]

    def get_compile_debugfile_args(self, rel_obj: str, pch: bool = False) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_compile_debugfile_args(rel_obj, pch))

    def get_link_debugfile_args(self, targetfile: str) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_link_debugfile_args(targetfile), Phase.LINKER)

    def get_depfile_suffix(self) -> str:
        return 'd'

    def get_optimization_link_args(self, optimization_level: str) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_optimization_link_args(optimization_level), Phase.LINKER)

    def build_rpath_args(self, env: Environment, build_dir: str, from_dir: str,
                         target: BuildTarget, extra_paths: T.Optional[T.List[str]] = None) -> T.Tuple[T.List[str], T.Set[bytes]]:
        (rpath_args, rpath_dirs_to_remove) = self.host_compiler.build_rpath_args(
            env, build_dir, from_dir, target, extra_paths)
        return (self._to_host_flags(rpath_args, Phase.LINKER), rpath_dirs_to_remove)

    def linker_to_compiler_args(self, args: T.List[str]) -> T.List[str]:
        return args

    def get_pic_args(self) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_pic_args())

    def compute_parameters_with_absolute_paths(self, parameter_list: T.List[str],
                                               build_dir: str) -> T.List[str]:
        return []

    def get_output_args(self, target: str) -> T.List[str]:
        return ['-o', target]

    def get_dependency_gen_args(self, outtarget: str, outfile: str) -> T.List[str]:
        if version_compare(self.version, '>= 10.2'):
            # According to nvcc Documentation, `-MD` option is added after 10.2
            # Reference: [CUDA 10.1](https://docs.nvidia.com/cuda/archive/10.1/cuda-compiler-driver-nvcc/index.html#options-for-specifying-compilation-phase-generate-nonsystem-dependencies)
            # Reference: [CUDA 10.2](https://docs.nvidia.com/cuda/archive/10.2/cuda-compiler-driver-nvcc/index.html#options-for-specifying-compilation-phase-generate-nonsystem-dependencies)
            return ['-MD', '-MT', outtarget, '-MF', outfile]
        else:
            return []

    def get_std_exe_link_args(self) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_std_exe_link_args(), Phase.LINKER)

    def find_library(self, libname: str, env: 'Environment', extra_dirs: T.List[str],
                     libtype: LibType = LibType.PREFER_SHARED, lib_prefix_warning: bool = True, ignore_system_dirs: bool = False) -> T.Optional[T.List[str]]:
        return self.host_compiler.find_library(libname, env, extra_dirs, libtype, lib_prefix_warning, ignore_system_dirs)

    def get_crt_compile_args(self, crt_val: str, buildtype: str) -> T.List[str]:
        return self._to_host_flags(self.host_compiler.get_crt_compile_args(crt_val, buildtype))

    def get_crt_link_args(self, crt_val: str, buildtype: str) -> T.List[str]:
        # nvcc defaults to static, release version of msvc runtime and provides no
        # native option to override it; override it with /NODEFAULTLIB
        host_link_arg_overrides = []
        host_crt_compile_args = self.host_compiler.get_crt_compile_args(crt_val, buildtype)
        if any(arg in {'/MDd', '/MD', '/MTd'} for arg in host_crt_compile_args):
            host_link_arg_overrides += ['/NODEFAULTLIB:LIBCMT.lib']
        return self._to_host_flags(host_link_arg_overrides + self.host_compiler.get_crt_link_args(crt_val, buildtype), Phase.LINKER)

    def get_target_link_args(self, target: 'BuildTarget') -> T.List[str]:
        return self._to_host_flags(super().get_target_link_args(target), Phase.LINKER)

    def get_dependency_compile_args(self, dep: 'Dependency') -> T.List[str]:
        return self._to_host_flags(super().get_dependency_compile_args(dep))

    def get_dependency_link_args(self, dep: 'Dependency') -> T.List[str]:
        return self._to_host_flags(super().get_dependency_link_args(dep), Phase.LINKER)

    def get_ccbin_args(self,
                       target: 'T.Optional[BuildTarget]',
                       env: 'Environment',
                       subproject: T.Optional[str] = None) -> T.List[str]:
        key = self.form_compileropt_key('ccbindir').evolve(subproject=subproject)
        if target:
            ccbindir = env.coredata.get_option_for_target(target, key)
        else:
            ccbindir = env.coredata.optstore.get_value_for(key)
        if isinstance(ccbindir, str) and ccbindir != '':
            return [self._shield_nvcc_list_arg('-ccbin='+ccbindir, False)]
        else:
            return []

    def get_profile_generate_args(self) -> T.List[str]:
        return ['-Xcompiler=' + x for x in self.host_compiler.get_profile_generate_args()]

    def get_profile_use_args(self) -> T.List[str]:
        return ['-Xcompiler=' + x for x in self.host_compiler.get_profile_use_args()]

    def get_assert_args(self, disable: bool, env: 'Environment') -> T.List[str]:
        cccl_macros = []
        if not disable and self.debug_macros_available:
            # https://github.com/NVIDIA/cccl/pull/2382
            cccl_macros = ['-DCCCL_ENABLE_ASSERTIONS=1']

        return self.host_compiler.get_assert_args(disable, env) + cccl_macros

    def has_multi_arguments(self, args: T.List[str], env: Environment) -> T.Tuple[bool, bool]:
        args = self._to_host_flags(args)
        return self.compiles('int main(void) { return 0; }', env, extra_args=args, mode=CompileCheckMode.COMPILE)

    def has_multi_link_arguments(self, args: T.List[str], env: Environment) -> T.Tuple[bool, bool]:
        args = ['-Xnvlink='+self._shield_nvcc_list_arg(s) for s in self.linker.fatal_warnings()]
        args += self._to_host_flags(args, phase=Phase.LINKER)
        return self.compiles('int main(void) { return 0; }', env, extra_args=args, mode=CompileCheckMode.LINK)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 The Meson development team

from __future__ import annotations
from collections import defaultdict
import os
import tempfile
import typing as T

from .run_tool import run_tool_on_targets, run_with_buffered_output
from .. import build, mlog
from ..mesonlib import MachineChoice, PerMachine

if T.TYPE_CHECKING:
    from ..compilers.rust import RustCompiler

class ClippyDriver:
    def __init__(self, build: build.Build, tempdir: str):
        self.tools: PerMachine[T.List[str]] = PerMachine([], [])
        self.warned: T.DefaultDict[str, bool] = defaultdict(lambda: False)
        self.tempdir = tempdir
        for machine in MachineChoice:
            compilers = build.environment.coredata.compilers[machine]
            if 'rust' in compilers:
                compiler = T.cast('RustCompiler', compilers['rust'])
                self.tools[machine] = compiler.get_rust_tool('clippy-driver', build.environment)

    def warn_missing_clippy(self, machine: str) -> None:
        if self.warned[machine]:
            return
        mlog.warning(f'clippy-driver not found for {machine} machine')
        self.warned[machine] = True

    def __call__(self, target: T.Dict[str, T.Any]) -> T.Iterable[T.Coroutine[None, None, int]]:
        for src_block in target['target_sources']:
            if 'compiler' in src_block and src_block['language'] == 'rust':
                clippy = getattr(self.tools, src_block['machine'])
                if not clippy:
                    self.warn_missing_clippy(src_block['machine'])
                    continue

                cmdlist = list(clippy)
                prev = None
                lints_cap = None
                for arg in src_block['parameters']:
                    if prev == '--cap-lints':
                        cmdlist.append(prev)
                        lints_cap = arg
                        prev = None
                    elif prev:
                        prev = None
                        continue
                    if arg in {'--emit', '--out-dir', '--cap-lints'}:
                        prev = arg
                    else:
                        cmdlist.append(arg)

                # no use in running clippy if it wouldn't print anything anyway
                if lints_cap == 'allow':
                    break

                cmdlist.extend(src_block['sources'])
                # the default for --emit is to go all the way to linking,
                # and --emit dep-info= is not enough for clippy to do
                # enough analysis, so use --emit metadata.
                cmdlist.append('--emit')
                cmdlist.append('metadata')
                cmdlist.append('--out-dir')
                cmdlist.append(self.tempdir)
                yield run_with_buffered_output(cmdlist)

def run(args: T.List[str]) -> int:
    os.chdir(args[0])
    build_data = build.load(os.getcwd())
    with tempfile.TemporaryDirectory() as d:
        return run_tool_on_targets(ClippyDriver(build_data, d))



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import datetime

import portage
from portage import os
from portage.repository.storage.interface import (
    RepoStorageException,
    RepoStorageInterface,
)
from portage.util.futures import asyncio

from _emerge.SpawnProcess import SpawnProcess


class HardlinkRcuRepoStorage(RepoStorageInterface):
    """
    Enable read-copy-update (RCU) behavior for sync operations. The
    current latest immutable version of a repository will be
    reference by a symlink found where the repository would normally
    be located.  Repository consumers should resolve the canonical
    path of this symlink before attempt to access the repository,
    and all operations should be read-only, since the repository
    is considered immutable. Updates occur by atomic replacement
    of the symlink, which causes new consumers to use the new
    immutable version, while any earlier consumers continue to use
    the canonical path that was resolved earlier.

    Performance is better than HardlinkQuarantineRepoStorage,
    since commit involves atomic replacement of a symlink. Since
    the symlink usage would require special handling for scenarios
    involving bind mounts and chroots, this module is not enabled
    by default.

    repos.conf parameters:

            sync-rcu-store-dir

                    Directory path reserved for sync-rcu storage. This
                    directory must have a unique value for each repository
                    (do not set it in the DEFAULT section).  This directory
                    must not contain any other files or directories aside
                    from those that are created automatically when sync-rcu
                    is enabled.

            sync-rcu-spare-snapshots = 1

                    Number of spare snapshots for sync-rcu to retain with
                    expired ttl. This protects the previous latest snapshot
                    from being removed immediately after a new version
                    becomes available, since it might still be used by
                    running processes.

            sync-rcu-ttl-days = 7

                    Number of days for sync-rcu to retain previous immutable
                    snapshots of a repository. After the ttl of a particular
                    snapshot has expired, it will be remove automatically (the
                    latest snapshot is exempt, and sync-rcu-spare-snapshots
                    configures the number of previous snapshots that are
                    exempt). If the ttl is set too low, then a snapshot could
                    expire while it is in use by a running process.

    """

    def __init__(self, repo, spawn_kwargs):
        # Note that repo.location cannot substitute for repo.user_location here,
        # since we manage a symlink that resides at repo.user_location, and
        # repo.location is the irreversible result of realpath(repo.user_location).
        self._user_location = repo.user_location
        self._spawn_kwargs = spawn_kwargs

        if not repo.sync_allow_hardlinks:
            raise RepoStorageException(
                "repos.conf sync-rcu setting"
                " for repo '%s' requires that sync-allow-hardlinks be enabled"
                % repo.name
            )

        # Raise an exception if repo.sync_rcu_store_dir is unset, since the
        # user needs to be aware of this location for bind mount and chroot
        # scenarios
        if not repo.sync_rcu_store_dir:
            raise RepoStorageException(
                "repos.conf sync-rcu setting"
                " for repo '%s' requires that sync-rcu-store-dir be set" % repo.name
            )

        self._storage_location = repo.sync_rcu_store_dir
        if repo.sync_rcu_spare_snapshots is None or repo.sync_rcu_spare_snapshots < 0:
            self._spare_snapshots = 1
        else:
            self._spare_snapshots = repo.sync_rcu_spare_snapshots
        if self._spare_snapshots < 0:
            self._spare_snapshots = 0
        if repo.sync_rcu_ttl_days is None or repo.sync_rcu_ttl_days < 0:
            self._ttl_days = 1
        else:
            self._ttl_days = repo.sync_rcu_ttl_days
        self._update_location = None
        self._latest_symlink = os.path.join(self._storage_location, "latest")
        self._latest_canonical = os.path.realpath(self._latest_symlink)
        if not os.path.exists(self._latest_canonical) or os.path.islink(
            self._latest_canonical
        ):
            # It doesn't exist, or it's a broken symlink.
            self._latest_canonical = None
        self._snapshots_dir = os.path.join(self._storage_location, "snapshots")

    async def _check_call(self, cmd, privileged=False):
        """
        Run cmd and raise RepoStorageException on failure.

        @param cmd: command to execute
        @type cmd: list
        @param privileged: run with maximum privileges
        @type privileged: bool
        """
        if privileged:
            kwargs = dict(fd_pipes=self._spawn_kwargs.get("fd_pipes"))
        else:
            kwargs = self._spawn_kwargs
        p = SpawnProcess(args=cmd, scheduler=asyncio.get_event_loop(), **kwargs)
        p.start()
        if await p.async_wait() != os.EX_OK:
            raise RepoStorageException(
                f"command exited with status {p.returncode}: {' '.join(cmd)}"
            )

    async def init_update(self):
        update_location = os.path.join(self._storage_location, "update")
        await self._check_call(["rm", "-rf", update_location])

        # This assumes normal umask permissions if it doesn't exist yet.
        portage.util.ensure_dirs(self._storage_location)

        if self._latest_canonical is not None:
            portage.util.ensure_dirs(update_location)
            portage.util.apply_stat_permissions(
                update_location, os.stat(self._user_location)
            )
            # Use  rsync --link-dest to hardlink a files into update_location,
            # since cp -l is not portable.
            await self._check_call(
                [
                    "rsync",
                    "-a",
                    "--link-dest",
                    self._latest_canonical,
                    self._latest_canonical + "/",
                    update_location + "/",
                ]
            )

        elif not os.path.islink(self._user_location):
            await self._migrate(update_location)
            update_location = await self.init_update()

        self._update_location = update_location

        return self._update_location

    async def _migrate(self, update_location):
        """
        When repo.user_location is a normal directory, migrate it to
        storage so that it can be replaced with a symlink. After migration,
        commit the content as the latest snapshot.
        """
        try:
            os.rename(self._user_location, update_location)
        except OSError:
            portage.util.ensure_dirs(update_location)
            portage.util.apply_stat_permissions(
                update_location, os.stat(self._user_location)
            )
            # It's probably on a different device, so copy it.
            await self._check_call(
                ["rsync", "-a", self._user_location + "/", update_location + "/"]
            )

            # Remove the old copy so that symlink can be created. Run with
            # maximum privileges, since removal requires write access to
            # the parent directory.
            await self._check_call(["rm", "-rf", self._user_location], privileged=True)

        self._update_location = update_location

        # Make this copy the latest snapshot
        await self.commit_update()

    @property
    def current_update(self):
        if self._update_location is None:
            raise RepoStorageException("current update does not exist")
        return self._update_location

    async def commit_update(self):
        update_location = self.current_update
        self._update_location = None
        try:
            snapshots = [int(name) for name in os.listdir(self._snapshots_dir)]
        except OSError:
            snapshots = []
            portage.util.ensure_dirs(self._snapshots_dir)
            portage.util.apply_stat_permissions(
                self._snapshots_dir, os.stat(self._storage_location)
            )
        if snapshots:
            new_id = max(snapshots) + 1
        else:
            new_id = 1
        os.rename(update_location, os.path.join(self._snapshots_dir, str(new_id)))
        new_symlink = self._latest_symlink + ".new"
        try:
            os.unlink(new_symlink)
        except OSError:
            pass
        os.symlink(f"snapshots/{new_id}", new_symlink)

        # If SyncManager.pre_sync creates an empty directory where
        # self._latest_symlink is suppose to be (which is normal if
        # sync-rcu-store-dir has been removed), then we need to remove
        # the directory or else rename will raise IsADirectoryError
        # when we try to replace the directory with a symlink.
        try:
            os.rmdir(self._latest_symlink)
        except OSError:
            pass

        os.rename(new_symlink, self._latest_symlink)

        try:
            user_location_correct = os.path.samefile(
                self._user_location, self._latest_symlink
            )
        except OSError:
            user_location_correct = False

        if not user_location_correct:
            new_symlink = self._user_location + ".new"
            try:
                os.unlink(new_symlink)
            except OSError:
                pass
            os.symlink(self._latest_symlink, new_symlink)
            os.rename(new_symlink, self._user_location)

    async def abort_update(self):
        if self._update_location is not None:
            update_location = self._update_location
            self._update_location = None
            await self._check_call(["rm", "-rf", update_location])

    async def garbage_collection(self):
        snap_ttl = datetime.timedelta(days=self._ttl_days)
        snapshots = sorted(int(name) for name in os.listdir(self._snapshots_dir))
        # always preserve the latest snapshot
        protect_count = self._spare_snapshots + 1
        while snapshots and protect_count:
            protect_count -= 1
            snapshots.pop()
        for snap_id in snapshots:
            snap_path = os.path.join(self._snapshots_dir, str(snap_id))
            try:
                st = os.stat(snap_path)
            except OSError:
                continue
            snap_timestamp = datetime.datetime.fromtimestamp(
                st.st_mtime, tz=datetime.timezone.utc
            )
            if (
                datetime.datetime.now(datetime.timezone.utc) - snap_timestamp
            ) < snap_ttl:
                continue
            await self._check_call(["rm", "-rf", snap_path])



# elog/messages.py - elog core functions
# Copyright 2006-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2


def filter_loglevels(logentries, loglevels):
    # remove unwanted entries from all logentries
    rValue = {}
    loglevels = [x.upper() for x in loglevels]
    for phase in logentries:
        for msgtype, msgcontent in logentries[phase]:
            if msgtype.upper() in loglevels or "*" in loglevels:
                if phase not in rValue:
                    rValue[phase] = []
                rValue[phase].append((msgtype, msgcontent))
    return rValue



# SPDX-License-Identifier: Apache-2.0
# Copyright 2019-2022 The meson development team

from __future__ import annotations

"""Abstractions for the LLVM/Clang compiler family."""

import os
import shutil
import typing as T

from ... import mesonlib
from ... import options
from ...linkers.linkers import AppleDynamicLinker, ClangClDynamicLinker, LLVMDynamicLinker, GnuGoldDynamicLinker, \
    MoldDynamicLinker, MSVCDynamicLinker
from ...options import OptionKey
from ..compilers import CompileCheckMode
from .gnu import GnuLikeCompiler

if T.TYPE_CHECKING:
    from ...options import MutableKeyedOptionDictType
    from ...environment import Environment
    from ...dependencies import Dependency  # noqa: F401
    from ..compilers import Compiler

    CompilerMixinBase = Compiler
else:
    CompilerMixinBase = object

clang_color_args: T.Dict[str, T.List[str]] = {
    'auto': ['-fdiagnostics-color=auto'],
    'always': ['-fdiagnostics-color=always'],
    'never': ['-fdiagnostics-color=never'],
}

clang_optimization_args: T.Dict[str, T.List[str]] = {
    'plain': [],
    '0': ['-O0'],
    'g': ['-Og'],
    '1': ['-O1'],
    '2': ['-O2'],
    '3': ['-O3'],
    's': ['-Oz'],
}

clang_lang_map = {
    'c': 'c',
    'cpp': 'c++',
    'objc': 'objective-c',
    'objcpp': 'objective-c++',
}

class ClangCompiler(GnuLikeCompiler):

    id = 'clang'

    def __init__(self, defines: T.Optional[T.Dict[str, str]]):
        super().__init__()
        self.defines = defines or {}
        self.base_options.update(
            {OptionKey('b_colorout'), OptionKey('b_lto_threads'), OptionKey('b_lto_mode'), OptionKey('b_thinlto_cache'),
             OptionKey('b_thinlto_cache_dir')})

        # TODO: this really should be part of the linker base_options, but
        # linkers don't have base_options.
        if isinstance(self.linker, AppleDynamicLinker):
            self.base_options.add(OptionKey('b_bitcode'))
        elif isinstance(self.linker, MSVCDynamicLinker):
            self.base_options.add(OptionKey('b_vscrt'))
        # All Clang backends can also do LLVM IR
        self.can_compile_suffixes.add('ll')

    def get_colorout_args(self, colortype: str) -> T.List[str]:
        return clang_color_args[colortype][:]

    def has_builtin_define(self, define: str) -> bool:
        return define in self.defines

    def get_builtin_define(self, define: str) -> T.Optional[str]:
        return self.defines.get(define)

    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
        return clang_optimization_args[optimization_level]

    def get_pch_suffix(self) -> str:
        return 'pch'

    def get_pch_use_args(self, pch_dir: str, header: str) -> T.List[str]:
        # Workaround for Clang bug http://llvm.org/bugs/show_bug.cgi?id=15136
        # This flag is internal to Clang (or at least not documented on the man page)
        # so it might change semantics at any time.
        return ['-include-pch', os.path.join(pch_dir, self.get_pch_name(header))]

    def get_compiler_check_args(self, mode: CompileCheckMode) -> T.List[str]:
        # Clang is different than GCC, it will return True when a symbol isn't
        # defined in a header. Specifically this is caused by a functionality
        # both GCC and clang have: for some "well known" functions, arbitrarily
        # chosen, they provide fixit suggestions for the header you should try
        # including.
        #
        # - With GCC, this is a note appended to the prexisting diagnostic
        #   "error: undeclared identifier"
        #
        # - With clang, the error is converted to a c89'ish implicit function
        #   declaration instead, which can be disabled with -Wno-error and on
        #   clang < 16, simply passes compilation by default.
        #
        # One example of a clang fixit suggestion is for `strlcat`, which
        # triggers this.
        #
        # This was reported in 2017 and promptly fixed. Just kidding!
        # https://github.com/llvm/llvm-project/issues/33905
        myargs: T.List[str] = ['-Werror=implicit-function-declaration']
        if mode is CompileCheckMode.COMPILE:
            myargs.extend(['-Werror=unknown-warning-option', '-Werror=unused-command-line-argument'])
            if mesonlib.version_compare(self.version, '>=3.6.0'):
                myargs.append('-Werror=ignored-optimization-argument')
        return super().get_compiler_check_args(mode) + myargs

    def has_function(self, funcname: str, prefix: str, env: 'Environment', *,
                     extra_args: T.Optional[T.List[str]] = None,
                     dependencies: T.Optional[T.List['Dependency']] = None) -> T.Tuple[bool, bool]:
        if extra_args is None:
            extra_args = []
        # Starting with XCode 8, we need to pass this to force linker
        # visibility to obey OS X/iOS/tvOS minimum version targets with
        # -mmacosx-version-min, -miphoneos-version-min, -mtvos-version-min etc.
        # https://github.com/Homebrew/homebrew-core/issues/3727
        # TODO: this really should be communicated by the linker
        if isinstance(self.linker, AppleDynamicLinker) and mesonlib.version_compare(self.version, '>=8.0'):
            extra_args.append('-Wl,-no_weak_imports')
        return super().has_function(funcname, prefix, env, extra_args=extra_args,
                                    dependencies=dependencies)

    def openmp_flags(self, env: Environment) -> T.List[str]:
        if mesonlib.version_compare(self.version, '>=3.8.0'):
            return ['-fopenmp']
        elif mesonlib.version_compare(self.version, '>=3.7.0'):
            return ['-fopenmp=libomp']
        else:
            # Shouldn't work, but it'll be checked explicitly in the OpenMP dependency.
            return []

    def gen_vs_module_defs_args(self, defsfile: str) -> T.List[str]:
        if isinstance(self.linker, (ClangClDynamicLinker, MSVCDynamicLinker)):
            # With MSVC, DLLs only export symbols that are explicitly exported,
            # so if a module defs file is specified, we use that to export symbols
            return ['-Wl,/DEF:' + defsfile]
        return super().gen_vs_module_defs_args(defsfile)

    @classmethod
    def use_linker_args(cls, linker: str, version: str) -> T.List[str]:
        # Clang additionally can use a linker specified as a path, which GCC
        # (and other gcc-like compilers) cannot. This is because clang (being
        # llvm based) is retargetable, while GCC is not.
        #

        # eld: Qualcomm's opensource embedded linker
        if linker == 'eld':
            return ['-fuse-ld=eld']
        # qcld: Qualcomm's deprecated linker
        if linker == 'qcld':
            return ['-fuse-ld=qcld']
        if linker == 'mold':
            return ['-fuse-ld=mold']

        if shutil.which(linker):
            if not shutil.which(linker):
                raise mesonlib.MesonException(
                    f'Cannot find linker {linker}.')
            return [f'-fuse-ld={linker}']
        return super().use_linker_args(linker, version)

    def get_has_func_attribute_extra_args(self, name: str) -> T.List[str]:
        # Clang only warns about unknown or ignored attributes, so force an
        # error.
        return ['-Werror=attributes']

    def get_coverage_link_args(self) -> T.List[str]:
        return ['--coverage']

    def get_lto_compile_args(self, *, threads: int = 0, mode: str = 'default') -> T.List[str]:
        args: T.List[str] = []
        if mode == 'thin':
            # ThinLTO requires the use of gold, lld, ld64, lld-link or mold 1.1+
            if isinstance(self.linker, (MoldDynamicLinker)):
                # https://github.com/rui314/mold/commit/46995bcfc3e3113133620bf16445c5f13cd76a18
                if not mesonlib.version_compare(self.linker.version, '>=1.1'):
                    raise mesonlib.MesonException("LLVM's ThinLTO requires mold 1.1+")
            elif not isinstance(self.linker, (AppleDynamicLinker, ClangClDynamicLinker, LLVMDynamicLinker, GnuGoldDynamicLinker)):
                raise mesonlib.MesonException(f"LLVM's ThinLTO only works with gold, lld, lld-link, ld64 or mold, not {self.linker.id}")
            args.append(f'-flto={mode}')
        else:
            assert mode == 'default', 'someone forgot to wire something up'
            args.extend(super().get_lto_compile_args(threads=threads))
        return args

    def linker_to_compiler_args(self, args: T.List[str]) -> T.List[str]:
        if isinstance(self.linker, (ClangClDynamicLinker, MSVCDynamicLinker)):
            return [flag if flag.startswith('-Wl,') or flag.startswith('-fuse-ld=') else f'-Wl,{flag}' for flag in args]
        else:
            return args

    def get_lto_link_args(self, *, threads: int = 0, mode: str = 'default',
                          thinlto_cache_dir: T.Optional[str] = None) -> T.List[str]:
        args = self.get_lto_compile_args(threads=threads, mode=mode)
        if mode == 'thin' and thinlto_cache_dir is not None:
            # We check for ThinLTO linker support above in get_lto_compile_args, and all of them support
            # get_thinlto_cache_args as well
            args.extend(self.linker.get_thinlto_cache_args(thinlto_cache_dir))
        # In clang -flto-jobs=0 means auto, and is the default if unspecified, just like in meson
        if threads > 0:
            if not mesonlib.version_compare(self.version, '>=4.0.0'):
                raise mesonlib.MesonException('clang support for LTO threads requires clang >=4.0')
            args.append(f'-flto-jobs={threads}')
        return args


class ClangCStds(CompilerMixinBase):

    """Mixin class for clang based compilers for setting C standards.

    This is used by both ClangCCompiler and ClangClCompiler, as they share
    the same versions
    """

    _C17_VERSION = '>=6.0.0'
    _C18_VERSION = '>=8.0.0'
    _C2X_VERSION = '>=9.0.0'
    _C23_VERSION = '>=18.0.0'
    _C2Y_VERSION = '>=19.0.0'

    def get_options(self) -> MutableKeyedOptionDictType:
        opts = super().get_options()
        stds = ['c89', 'c99', 'c11']
        # https://releases.llvm.org/6.0.0/tools/clang/docs/ReleaseNotes.html
        # https://en.wikipedia.org/wiki/Xcode#Latest_versions
        if mesonlib.version_compare(self.version, self._C17_VERSION):
            stds += ['c17']
        if mesonlib.version_compare(self.version, self._C18_VERSION):
            stds += ['c18']
        if mesonlib.version_compare(self.version, self._C2X_VERSION):
            stds += ['c2x']
        if mesonlib.version_compare(self.version, self._C23_VERSION):
            stds += ['c23']
        if mesonlib.version_compare(self.version, self._C2Y_VERSION):
            stds += ['c2y']
        key = self.form_compileropt_key('std')
        std_opt = opts[key]
        assert isinstance(std_opt, options.UserStdOption), 'for mypy'
        std_opt.set_versions(stds, gnu=True)
        return opts


class ClangCPPStds(CompilerMixinBase):

    """Mixin class for clang based compilers for setting C++ standards.

    This is used by the ClangCPPCompiler
    """

    _CPP23_VERSION = '>=12.0.0'
    _CPP26_VERSION = '>=17.0.0'

    def get_options(self) -> MutableKeyedOptionDictType:
        opts = super().get_options()
        stds = [
            'c++98', 'c++03', 'c++11', 'c++14', 'c++17', 'c++1z', 'c++2a',
            'c++20',
        ]
        if mesonlib.version_compare(self.version, self._CPP23_VERSION):
            stds.append('c++23')
        if mesonlib.version_compare(self.version, self._CPP26_VERSION):
            stds.append('c++26')
        key = self.form_compileropt_key('std')
        std_opt = opts[key]
        assert isinstance(std_opt, options.UserStdOption), 'for mypy'
        std_opt.set_versions(stds, gnu=True)
        return opts



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

import typing


@typing.final
class InfinityType:
    __slots__ = ()

    def __repr__(self) -> str:
        return "Infinity"

    def __hash__(self) -> int:
        return hash(repr(self))

    def __lt__(self, other: object) -> bool:
        return False

    def __le__(self, other: object) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __gt__(self, other: object) -> bool:
        return True

    def __ge__(self, other: object) -> bool:
        return True

    def __neg__(self: object) -> "NegativeInfinityType":
        return NegativeInfinity


Infinity = InfinityType()


@typing.final
class NegativeInfinityType:
    __slots__ = ()

    def __repr__(self) -> str:
        return "-Infinity"

    def __hash__(self) -> int:
        return hash(repr(self))

    def __lt__(self, other: object) -> bool:
        return True

    def __le__(self, other: object) -> bool:
        return True

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __gt__(self, other: object) -> bool:
        return False

    def __ge__(self, other: object) -> bool:
        return False

    def __neg__(self: object) -> InfinityType:
        return Infinity


NegativeInfinity = NegativeInfinityType()



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2


class RootConfig:
    """This is used internally by depgraph to track information about a
    particular $ROOT."""

    __slots__ = ("mtimedb", "root", "setconfig", "sets", "settings", "trees")

    pkg_tree_map = {"ebuild": "porttree", "binary": "bintree", "installed": "vartree"}

    tree_pkg_map = {}
    for k, v in pkg_tree_map.items():
        tree_pkg_map[v] = k

    def __init__(self, settings, trees, setconfig):
        self.trees = trees
        self.settings = settings
        self.root = self.settings["EROOT"]
        self.setconfig = setconfig
        if setconfig is None:
            self.sets = {}
        else:
            self.sets = self.setconfig.getSets()

    def update(self, other):
        """
        Shallow copy all attributes from another instance.
        """
        for k in self.__slots__:
            try:
                setattr(self, k, getattr(other, k))
            except AttributeError:
                # mtimedb is currently not a required attribute
                try:
                    delattr(self, k)
                except AttributeError:
                    pass



# Copyright 1999-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import itertools
import logging
import re
import stat
import warnings

from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode
from portage.exception import (
    DigestException,
    FileNotFound,
    InvalidDataType,
    MissingParameter,
    PermissionDenied,
    PortageException,
    PortagePackageException,
)
from portage.const import MANIFEST2_HASH_DEFAULTS, MANIFEST2_IDENTIFIERS
from portage.localization import _

_manifest_re = re.compile(
    r"^(" + "|".join(MANIFEST2_IDENTIFIERS) + r") (\S+)( \d+( \S+ \S+)+)$", re.UNICODE
)


class FileNotInManifestException(PortageException):
    pass


def manifest2AuxfileFilter(filename):
    filename = filename.strip(os.sep)
    mysplit = filename.split(os.path.sep)
    if "CVS" in mysplit:
        return False
    for x in mysplit:
        if x[:1] == ".":
            return False
    return not filename[:7] == "digest-"


def manifest2MiscfileFilter(filename):
    return not (filename == "Manifest" or filename.endswith(".ebuild"))


def guessManifestFileType(filename):
    """Perform a best effort guess of which type the given filename is, avoid using this if possible"""
    if filename.startswith("files" + os.sep + "digest-"):
        return None
    if filename.startswith("files" + os.sep):
        return "AUX"
    if filename.endswith(".ebuild"):
        return "EBUILD"
    if filename in ["ChangeLog", "metadata.xml"]:
        return "MISC"
    return "DIST"


def guessThinManifestFileType(filename):
    filetype = guessManifestFileType(filename)
    if filetype != "DIST":
        return None
    return "DIST"


def parseManifest2(line):
    if not isinstance(line, str):
        line = " ".join(line)
    myentry = None
    matched = _manifest_re.match(line)
    if matched:
        tokens = matched.group(3).split()
        hashes = dict(zip(tokens[1::2], tokens[2::2]))
        hashes["size"] = int(tokens[0])
        myentry = Manifest2Entry(
            type=matched.group(1), name=matched.group(2), hashes=hashes
        )
    return myentry


class ManifestEntry:
    __slots__ = ("type", "name", "hashes")

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class Manifest2Entry(ManifestEntry):
    def __str__(self):
        myline = " ".join([self.type, self.name, str(self.hashes["size"])])
        myhashkeys = list(self.hashes)
        myhashkeys.remove("size")
        myhashkeys.sort()
        with_hashes = " ".join(f"{h} {self.hashes[h]}" for h in myhashkeys)
        return f"{myline} {with_hashes}"

    def __eq__(self, other):
        return (
            isinstance(other, Manifest2Entry)
            and self.type == other.type
            and self.name == other.name
            and self.hashes == other.hashes
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class Manifest:
    parsers = (parseManifest2,)

    def __init__(
        self,
        pkgdir,
        distdir=None,
        fetchlist_dict=None,
        manifest1_compat=DeprecationWarning,
        from_scratch=False,
        thin=False,
        allow_missing=False,
        allow_create=True,
        hashes=None,
        required_hashes=None,
        find_invalid_path_char=None,
        strict_misc_digests=True,
    ):
        """Create new Manifest instance for package in pkgdir.
        Do not parse Manifest file if from_scratch == True (only for internal use)
            The fetchlist_dict parameter is required only for generation of
            a Manifest (not needed for parsing and checking sums).
            If thin is specified, then the manifest carries only info for
            distfiles."""
        from portage.checksum import get_valid_checksum_keys
        from portage.repository.config import _find_invalid_path_char

        if manifest1_compat is not DeprecationWarning:
            warnings.warn(
                "The manifest1_compat parameter of the "
                "portage.manifest.Manifest constructor is deprecated.",
                DeprecationWarning,
                stacklevel=2,
            )

        if find_invalid_path_char is None:
            find_invalid_path_char = _find_invalid_path_char
        self._find_invalid_path_char = find_invalid_path_char
        self.pkgdir = _unicode_decode(pkgdir).rstrip(os.sep) + os.sep
        self.hashes = set()
        self.required_hashes = set()

        if hashes is None:
            hashes = MANIFEST2_HASH_DEFAULTS
        if required_hashes is None:
            required_hashes = hashes

        self.hashes.update(hashes)
        self.hashes.difference_update(
            hashname
            for hashname in list(self.hashes)
            if hashname not in get_valid_checksum_keys()
        )
        self.hashes.add("size")

        self.required_hashes.update(required_hashes)
        self.required_hashes.intersection_update(self.hashes)

        self.fhashdict = {t: {} for t in MANIFEST2_IDENTIFIERS}

        if not from_scratch:
            # Parse Manifest file for this instance
            try:
                self._readManifest(self.getFullname(), myhashdict=self.fhashdict)
            except FileNotFound:
                pass

        self.fetchlist_dict = {}
        if fetchlist_dict:
            self.fetchlist_dict.update(fetchlist_dict)

        self.distdir = distdir
        self.thin = thin
        if thin:
            self.guessType = guessThinManifestFileType
        else:
            self.guessType = guessManifestFileType
        self.allow_missing = allow_missing
        self.allow_create = allow_create
        self.strict_misc_digests = strict_misc_digests

    def getFullname(self):
        """Returns the absolute path to the Manifest file for this instance"""
        return os.path.join(self.pkgdir, "Manifest")

    def getDigests(self):
        """Compability function for old digest/manifest code, returns dict of filename:{hashfunction:hashvalue}"""
        rval = {
            k: v for t in MANIFEST2_IDENTIFIERS for k, v in self.fhashdict[t].items()
        }
        return rval

    def getTypeDigests(self, ftype):
        """Similar to getDigests(), but restricted to files of the given type."""
        return self.fhashdict[ftype]

    def _readManifest(self, file_path, myhashdict=None, **kwargs):
        """Parse a manifest.  If myhashdict is given then data will be added too it.
        Otherwise, a new dict will be created and returned."""
        try:
            with open(
                _unicode_encode(file_path, encoding=_encodings["fs"], errors="strict"),
                encoding=_encodings["repo.content"],
                errors="replace",
            ) as f:
                if myhashdict is None:
                    myhashdict = {}
                self._parseDigests(f, myhashdict=myhashdict, **kwargs)
            return myhashdict
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise FileNotFound(file_path)
            else:
                raise

    def _parseManifestLines(self, mylines):
        """Parse manifest lines and return a list of manifest entries."""
        for myline in mylines:
            myentry = None
            for parser in self.parsers:
                myentry = parser(myline)
                if myentry is not None:
                    yield myentry
                    break  # go to the next line

    def _parseDigests(self, mylines, myhashdict=None, mytype=None):
        """Parse manifest entries and store the data in myhashdict.  If mytype
        is specified, it will override the type for all parsed entries."""
        if myhashdict is None:
            myhashdict = {}
        for myentry in self._parseManifestLines(mylines):
            if mytype is None:
                myentry_type = myentry.type
            else:
                myentry_type = mytype
            myhashdict.setdefault(myentry_type, {})
            myhashdict[myentry_type].setdefault(myentry.name, {})
            myhashdict[myentry_type][myentry.name].update(myentry.hashes)
        return myhashdict

    def _getDigestData(self, distlist):
        """create a hash dict for a specific list of files"""
        myhashdict = {
            mytype: {myname: self.fhashdict[mytype][myname]}
            for myname in distlist
            for mytype in self.fhashdict
            if myname in self.fhashdict[mytype]
        }
        return myhashdict

    def _createManifestEntries(self):
        from portage.checksum import get_valid_checksum_keys

        valid_hashes = set(itertools.chain(get_valid_checksum_keys(), ("size",)))
        mytypes = sorted(self.fhashdict)
        for mytype in mytypes:
            myfiles = sorted(self.fhashdict[mytype])
            for myfile in myfiles:
                remainings = set(self.fhashdict[mytype][myfile]).intersection(
                    valid_hashes
                )
                yield Manifest2Entry(
                    type=mytype,
                    name=myfile,
                    hashes={
                        remaining: self.fhashdict[mytype][myfile][remaining]
                        for remaining in remainings
                    },
                )

    def checkIntegrity(self):
        manifest_data = (
            (
                self.required_hashes.difference(set(self.fhashdict[mytype][myfile])),
                mytype,
                myfile,
            )
            for mytype in self.fhashdict
            for myfile in self.fhashdict[mytype]
        )
        for needed_hashes, its_type, its_file in manifest_data:
            if needed_hashes:
                raise MissingParameter(
                    _(
                        f"Missing {' '.join(needed_hashes)} checksum(s): {its_type} {its_file}"
                    )
                )

    def write(self, sign=False, force=False):
        """Write Manifest instance to disk, optionally signing it. Returns
        True if the Manifest is actually written, and False if the write
        is skipped due to existing Manifest being identical."""
        from portage.util import write_atomic

        rval = False
        if not self.allow_create:
            return rval
        self.checkIntegrity()
        try:
            myentries = list(self._createManifestEntries())
            update_manifest = True
            preserved_stats = {self.pkgdir.rstrip(os.sep): os.stat(self.pkgdir)}
            if myentries and not force:
                try:
                    with open(
                        _unicode_encode(
                            self.getFullname(),
                            encoding=_encodings["fs"],
                            errors="strict",
                        ),
                        encoding=_encodings["repo.content"],
                        errors="replace",
                    ) as f:
                        oldentries = list(self._parseManifestLines(f))
                        preserved_stats[self.getFullname()] = os.fstat(f.fileno())
                        if len(oldentries) == len(myentries):
                            update_manifest = False
                            for oldentry, myentry in zip(oldentries, myentries):
                                if oldentry != myentry:
                                    update_manifest = True
                                    break
                except OSError as e:
                    if e.errno == errno.ENOENT:
                        pass
                    else:
                        raise

            if update_manifest:
                if myentries or not (self.thin or self.allow_missing):
                    # If myentries is empty, don't write an empty manifest
                    # when thin or allow_missing is enabled. Except for
                    # thin manifests with no DIST entries, myentries is
                    # non-empty for all currently known use cases.
                    write_atomic(
                        self.getFullname(),
                        "".join(f"{myentry}\n" for myentry in myentries),
                    )
                    self._apply_max_mtime(preserved_stats, myentries)
                    rval = True
                else:
                    # With thin manifest, there's no need to have
                    # a Manifest file if there are no DIST entries.
                    try:
                        os.unlink(self.getFullname())
                    except OSError as e:
                        if e.errno != errno.ENOENT:
                            raise
                    rval = True

            if sign:
                self.sign()
        except OSError as e:
            if e.errno == errno.EACCES:
                raise PermissionDenied(str(e))
            raise
        return rval

    def _apply_max_mtime(self, preserved_stats, entries):
        """
        Set the Manifest mtime to the max mtime of all relevant files
        and directories. Directory mtimes account for file renames and
        removals. The existing Manifest mtime accounts for eclass
        modifications that change DIST entries. This results in a
        stable/predictable mtime, which is useful when converting thin
        manifests to thick manifests for distribution via rsync. For
        portability, the mtime is set with 1 second resolution.

        @param preserved_stats: maps paths to preserved stat results
                that should be used instead of os.stat() calls
        @type preserved_stats: dict
        @param entries: list of current Manifest2Entry instances
        @type entries: list
        """
        from portage.util import writemsg_level

        # Use stat_result[stat.ST_MTIME] for 1 second resolution, since
        # it always rounds down. Note that stat_result.st_mtime will round
        # up from 0.999999999 to 1.0 when precision is lost during conversion
        # from nanosecond resolution to float.

        def _update_max(max_mtime, st):
            stat_mtime = st[stat.ST_MTIME]
            if max_mtime:
                return max(max_mtime, stat_mtime)

        def _stat(path):
            if path in preserved_stats:
                return preserved_stats[path]
            else:
                return os.stat(path)

        max_mtime = None
        for stat_result in preserved_stats.values():
            max_mtime = _update_max(max_mtime, stat_result)

        for entry in entries:
            if entry.type == "DIST":
                continue
            files = ""
            if entry.type == "AUX":
                files = "files"
            abs_path = os.path.join(self.pkgdir, files, entry.name)
            max_mtime = _update_max(max_mtime, _stat(abs_path))

        if not self.thin:
            # Account for changes to all relevant nested directories.
            # This is not necessary for thin manifests because
            # self.pkgdir is already included via preserved_stats.
            for parent_dir, dirs, files in os.walk(self.pkgdir.rstrip(os.sep)):
                try:
                    parent_dir = _unicode_decode(
                        parent_dir, encoding=_encodings["fs"], errors="strict"
                    )
                except UnicodeDecodeError:
                    # If an absolute path cannot be decoded, then it is
                    # always excluded from the manifest (repoman will
                    # report such problems).
                    pass
                else:
                    max_mtime = _update_max(max_mtime, _stat(parent_dir))

        if max_mtime is not None:
            for path in preserved_stats:
                try:
                    os.utime(path, (max_mtime, max_mtime))
                except OSError as e:
                    # Even though we have write permission, utime fails
                    # with EPERM if path is owned by a different user.
                    # Only warn in this case, since it's not a problem
                    # unless this repo is being prepared for distribution
                    # via rsync.
                    writemsg_level(
                        f"!!! utime('{path}', ({max_mtime}, {max_mtime})): {e}\n",
                        level=logging.WARNING,
                        noiselevel=-1,
                    )

    def sign(self):
        """Sign the Manifest"""
        raise NotImplementedError()

    def validateSignature(self):
        """Validate signature on Manifest"""
        raise NotImplementedError()

    def addFile(self, ftype, fname, hashdict=None, ignoreMissing=False):
        """Add entry to Manifest optionally using hashdict to avoid recalculation of hashes"""
        if ftype == "AUX" and not fname.startswith("files/"):
            fname = os.path.join("files", fname)
        if not os.path.exists(f"{self.pkgdir}{fname}") and not ignoreMissing:
            raise FileNotFound(fname)
        if ftype not in MANIFEST2_IDENTIFIERS:
            raise InvalidDataType(ftype)

        if fname.startswith("files"):
            fname = fname[6:]
        self.fhashdict[ftype][fname] = {}
        if hashdict is not None:
            self.fhashdict[ftype][fname].update(hashdict)
        if self.required_hashes.difference(set(self.fhashdict[ftype][fname])):
            self.updateAllFileHashes(
                ftype, [fname], checkExisting=False, ignoreMissing=ignoreMissing
            )

    def removeFile(self, ftype, fname):
        """Remove given entry from Manifest"""
        del self.fhashdict[ftype][fname]

    def hasFile(self, ftype, fname):
        """Return whether the Manifest contains an entry for the given type,filename pair"""
        return fname in self.fhashdict[ftype]

    def findFile(self, fname):
        """Return entrytype of the given file if present in Manifest or None if not present"""
        found_entries = (t for t in MANIFEST2_IDENTIFIERS if fname in self.fhashdict[t])
        return next(found_entries, None)

    def create(
        self,
        checkExisting=False,
        assumeDistHashesSometimes=False,
        assumeDistHashesAlways=False,
        requiredDistfiles=None,
    ):
        """Recreate this Manifest from scratch.  This will not use any
        existing checksums unless assumeDistHashesSometimes or
        assumeDistHashesAlways is true (assumeDistHashesSometimes will only
        cause DIST checksums to be reused if the file doesn't exist in
        DISTDIR).  The requiredDistfiles parameter specifies a list of
        distfiles to raise a FileNotFound exception for (if no file or existing
        checksums are available), and defaults to all distfiles when not
        specified."""
        from portage.checksum import perform_multiple_checksums

        if not self.allow_create:
            return
        if checkExisting:
            self.checkAllHashes()
        distfilehashes = {}
        if assumeDistHashesSometimes or assumeDistHashesAlways:
            distfilehashes.update(self.fhashdict["DIST"])
        self.__init__(
            self.pkgdir,
            distdir=self.distdir,
            fetchlist_dict=self.fetchlist_dict,
            from_scratch=True,
            thin=self.thin,
            allow_missing=self.allow_missing,
            allow_create=self.allow_create,
            hashes=self.hashes,
            required_hashes=self.required_hashes,
            find_invalid_path_char=self._find_invalid_path_char,
            strict_misc_digests=self.strict_misc_digests,
        )

        update_pkgdir = self._update_thick_pkgdir
        if self.thin:
            update_pkgdir = self._update_thin_pkgdir

        cpvlist = update_pkgdir(
            self._pkgdir_category(),
            os.path.basename(self.pkgdir.rstrip(os.path.sep)),
            self.pkgdir,
        )
        distlist = {
            distfile for cpv in cpvlist for distfile in self._getCpvDistfiles(cpv)
        }

        if requiredDistfiles is None:
            # This allows us to force removal of stale digests for the
            # ebuild --force digest option (no distfiles are required).
            requiredDistfiles = set()
        elif len(requiredDistfiles) == 0:
            # repoman passes in an empty list, which implies that all distfiles
            # are required.
            requiredDistfiles = distlist.copy()
        required_hash_types = set(itertools.chain(self.required_hashes, ("size",)))
        for f in distlist:
            fname = os.path.join(self.distdir, f)
            mystat = None
            try:
                mystat = os.stat(fname)
            except OSError:
                pass
            if (
                f in distfilehashes
                and not required_hash_types.difference(distfilehashes[f])
                and (
                    (assumeDistHashesSometimes and mystat is None)
                    or (assumeDistHashesAlways and mystat is None)
                    or (
                        assumeDistHashesAlways
                        and mystat is not None
                        and set(distfilehashes[f]) == set(self.hashes)
                        and distfilehashes[f]["size"] == mystat.st_size
                    )
                )
            ):
                self.fhashdict["DIST"][f] = distfilehashes[f]
            else:
                try:
                    self.fhashdict["DIST"][f] = perform_multiple_checksums(
                        fname, self.hashes
                    )
                except FileNotFound:
                    if f in requiredDistfiles:
                        raise

    def _is_cpv(self, cat, pn, filename):
        from portage.versions import _pkgsplit

        if not filename.endswith(".ebuild"):
            return None
        pf = filename[:-7]
        ps = _pkgsplit(pf)
        cpv = f"{cat}/{pf}"
        if not ps:
            raise PortagePackageException(_(f"Invalid package name: '{cpv}'"))
        if ps[0] != pn:
            raise PortagePackageException(
                _(f"Package name does not match directory name: '{cpv}'")
            )
        return cpv

    def _update_thin_pkgdir(self, cat, pn, pkgdir):
        _, _, pkgdir_files = next(os.walk(pkgdir), (None, None, None))

        def _process_for_cpv(filename):
            try:
                filename = _unicode_decode(
                    filename, encoding=_encodings["fs"], errors="strict"
                )
            except UnicodeDecodeError:
                return None
            if filename.startswith("."):
                return None
            pf = self._is_cpv(cat, pn, filename)
            if pf is not None:
                return pf

        processed = (_process_for_cpv(filename) for filename in pkgdir_files)
        cpvlist = [pf for pf in processed if pf]
        return cpvlist

    def _update_thick_pkgdir(self, cat, pn, pkgdir):
        from portage.checksum import perform_multiple_checksums

        _, _, pkgdir_files = next(os.walk(pkgdir), (None, None, None))
        cpvlist = []
        for f in pkgdir_files:
            try:
                f = _unicode_decode(f, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError:
                continue
            if f.startswith("."):
                continue
            pf = self._is_cpv(cat, pn, f)
            if pf is not None:
                mytype = "EBUILD"
                cpvlist.append(pf)
            elif self._find_invalid_path_char(f) == -1 and manifest2MiscfileFilter(f):
                mytype = "MISC"
            else:
                continue
            self.fhashdict[mytype][f] = perform_multiple_checksums(
                f"{self.pkgdir}{f}", self.hashes
            )
        recursive_files = []

        pkgdir = self.pkgdir
        cut_len = len(os.path.join(pkgdir, f"files{os.sep}"))
        for parentdir, dirs, files in os.walk(os.path.join(pkgdir, "files")):
            for f in files:
                try:
                    f = _unicode_decode(f, encoding=_encodings["fs"], errors="strict")
                except UnicodeDecodeError:
                    continue
                full_path = os.path.join(parentdir, f)
                recursive_files.append(full_path[cut_len:])
        for f in recursive_files:
            if self._find_invalid_path_char(f) != -1 or not manifest2AuxfileFilter(f):
                continue
            self.fhashdict["AUX"][f] = perform_multiple_checksums(
                os.path.join(self.pkgdir, "files", f.lstrip(os.sep)), self.hashes
            )
        return cpvlist

    def _pkgdir_category(self):
        return self.pkgdir.rstrip(os.sep).split(os.sep)[-2]

    def _getAbsname(self, ftype, fname):
        if ftype == "DIST":
            abspath = (self.distdir, fname)
        elif ftype == "AUX":
            abspath = (self.pkgdir, "files", fname)
        else:
            abspath = (self.pkgdir, fname)
        return os.path.join(*abspath)

    def checkAllHashes(self, ignoreMissingFiles=False):
        for t in MANIFEST2_IDENTIFIERS:
            self.checkTypeHashes(t, ignoreMissingFiles=ignoreMissingFiles)

    def checkTypeHashes(self, idtype, ignoreMissingFiles=False, hash_filter=None):
        for f in self.fhashdict[idtype]:
            self.checkFileHashes(
                idtype, f, ignoreMissing=ignoreMissingFiles, hash_filter=hash_filter
            )

    def checkFileHashes(self, ftype, fname, ignoreMissing=False, hash_filter=None):
        from portage.checksum import (
            _apply_hash_filter,
            _filter_unaccelarated_hashes,
            verify_all,
        )

        digests = _filter_unaccelarated_hashes(self.fhashdict[ftype][fname])
        if hash_filter is not None:
            digests = _apply_hash_filter(digests, hash_filter)
        try:
            ok, reason = verify_all(self._getAbsname(ftype, fname), digests)
            if not ok:
                raise DigestException(
                    tuple([self._getAbsname(ftype, fname)] + list(reason))
                )
            return ok, reason
        except FileNotFound as e:
            if not ignoreMissing:
                raise
            return False, _(f"File Not Found: '{e}'")

    def checkCpvHashes(
        self, cpv, checkDistfiles=True, onlyDistfiles=False, checkMiscfiles=False
    ):
        """check the hashes for all files associated to the given cpv, include all
        AUX files and optionally all MISC files."""
        if not onlyDistfiles:
            self.checkTypeHashes("AUX", ignoreMissingFiles=False)
            if checkMiscfiles:
                self.checkTypeHashes("MISC", ignoreMissingFiles=False)
            ebuildname = f"{self._catsplit(cpv)[1]}.ebuild"
            self.checkFileHashes("EBUILD", ebuildname, ignoreMissing=False)
        if checkDistfiles or onlyDistfiles:
            for f in self._getCpvDistfiles(cpv):
                self.checkFileHashes("DIST", f, ignoreMissing=False)

    def _getCpvDistfiles(self, cpv):
        """Get a list of all DIST files associated to the given cpv"""
        return self.fetchlist_dict[cpv]

    def getDistfilesSize(self, fetchlist):
        total_bytes = sum(int(self.fhashdict["DIST"][f]["size"]) for f in fetchlist)
        return total_bytes

    def updateAllFileHashes(
        self, ftype, fnames, checkExisting=True, ignoreMissing=True, reuseExisting=False
    ):
        """Regenerate hashes from a list of files"""
        from portage.checksum import perform_multiple_checksums

        for fname in fnames:
            if checkExisting:
                self.checkFileHashes(ftype, fname, ignoreMissing=ignoreMissing)
            if not ignoreMissing and fname not in self.fhashdict[ftype]:
                raise FileNotInManifestException(fname)
            if fname not in self.fhashdict[ftype]:
                self.fhashdict[ftype][fname] = {}
            myhashkeys = self.hashes
            if reuseExisting:
                myhashkeys = myhashkeys.difference(self.fhashdict[ftype][fname])
            myhashes = perform_multiple_checksums(
                self._getAbsname(ftype, fname), myhashkeys
            )
            self.fhashdict[ftype][fname].update(myhashes)

    def updateAllTypeHashes(
        self, idtypes, checkExisting=False, ignoreMissingFiles=True
    ):
        """Regenerate all hashes for all files from a list of types"""
        for idtype in idtypes:
            self.updateAllFileHashes(
                ftype=idtype, fnames=self.fhashdict[idtype], checkExisting=checkExisting
            )

    def updateAllHashes(self, checkExisting=False, ignoreMissingFiles=True):
        """Regenerate all hashes for all files in this Manifest."""
        self.updateTypeHashes(
            idtypes=MANIFEST2_IDENTIFIERS,
            checkExisting=checkExisting,
            ignoreMissingFiles=ignoreMissingFiles,
        )

    def updateCpvHashes(self, cpv, ignoreMissingFiles=True):
        """Regenerate all hashes associated to the given cpv (includes all AUX and MISC
        files)."""
        self.updateAllTypeHashes(
            idtypes=("AUX", "MISC"),
            ignoreMissingFiles=ignoreMissingFiles,
        )
        self.updateAllFileHashes(
            ftype="EBUILD",
            fnames=(f"{self._catsplit(cpv)[1]}.ebuild",),
            ignoreMissingFiles=ignoreMissingFiles,
        )
        self.updateAllFileHashes(
            ftype="DIST",
            fnames=self._getCpvDistfiles(cpv),
            ignoreMissingFiles=ignoreMissingFiles,
        )

    def updateHashesGuessType(self, fname, *args, **kwargs):
        """Regenerate hashes for the given file (guesses the type and then
        calls updateAllFileHashes)."""
        mytype = self.guessType(fname)
        if mytype is None:
            return
        elif mytype == "AUX":
            fname = fname[len(f"files{os.sep}") :]
        myrealtype = self.findFile(fname)
        if myrealtype is not None:
            mytype = myrealtype
        return self.updateAllFileHashes(ftype=mytype, fnames=(fname,), *args, **kwargs)

    def getFileData(self, ftype, fname, key):
        """Return the value of a specific (type,filename,key) triple, mainly useful
        to get the size for distfiles."""
        return self.fhashdict[ftype][fname][key]

    def getVersions(self):
        """Returns a list of manifest versions present in the manifest file."""
        mfname = self.getFullname()
        if not os.path.exists(mfname):
            return []
        with open(
            _unicode_encode(mfname, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["repo.content"],
            errors="replace",
        ) as myfile:
            line_splits = (line.split() for line in myfile.readlines())
            validation = (
                True
                for line_split in line_splits
                if len(line_split) > 4
                and line_split[0] in MANIFEST2_IDENTIFIERS
                and (len(line_split) - 3) % 2 == 0
            )
            if any(validation):
                return [2]
        return []

    def _catsplit(self, pkg_key):
        """Split a category and package, returning a list of [cat, pkg].
        This is compatible with portage.catsplit()"""
        return pkg_key.split("/", 1)
